from siliconcompiler.tools._common import input_provides, get_tool_task
from siliconcompiler.tools.xdm import setup as tool_setup
import os
import shutil


def setup(chip):
    tool_setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    if f'{design}.cir' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.cir',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,netlist,spice',
                 step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', f'{design}.xyce', step=step, index=index)

    chip.add('tool', tool, 'task', task, 'option', '--auto',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'option', '--source_file_format hspice',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'option', f'--dir_out outputs/{design}.xyce',
             step=step, index=index)

    chip.set('tool', 'xdm', 'task', 'convert', 'var', 'rename', 'true',
             step=step, index=index, clobber=False)
    chip.set('tool', 'xdm', 'task', 'convert', 'var', 'rename',
             'true/false: indicate whether to rename the output file to match the '
             'naming scheme for siliconcompiler', field='help')


def runtime_options(chip):
    return __get_input_file(chip)


def __get_input_file(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    design = chip.top()

    if f'{design}.cir' in input_provides(chip, step, index):
        return [f'inputs/{design}.cir']

    return chip.find_files('input', 'netlist', 'spice',
                           step=step, index=index)


def post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    if chip.get('tool', 'xdm', 'task', 'convert', 'var', 'rename',
                step=step, index=index) == ['false']:
        return

    inputfile = __get_input_file(chip)[0]

    outdir = f'outputs/{chip.top()}.xyce'

    inputfile_base = os.path.basename(inputfile)
    outputfile_base = f'{chip.top()}.cir'

    if inputfile_base != outputfile_base:
        shutil.move(os.path.join(outdir, inputfile_base), os.path.join(outdir, outputfile_base))




from siliconcompiler.tools.yosys.yosys import syn_setup, syn_post_process
import os
import re
import siliconcompiler.tools.yosys.prepareLib as prepareLib
from siliconcompiler import sc_open
from siliconcompiler import utils
from siliconcompiler.tools._common.asic import set_tool_task_var, get_libraries, get_mainlib
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.targets import asap7_demo


def make_docs(chip):
    chip.use(asap7_demo)


def setup(chip):
    '''
    Perform ASIC synthesis
    '''

    # Generic synthesis task setup.
    syn_setup(chip)

    # ASIC-specific setup.
    setup_asic(chip)


def setup_asic(chip):
    ''' Helper method for configs specific to ASIC steps (both syn and lec).
    '''

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['asic', 'logiclib']),
             step=step, index=index)

    syn_corners = get_synthesis_corner(chip)

    if syn_corners is None:
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'synthesis_corner']),
                 step=step, index=index)

    if syn_corners is not None:
        # add timing library requirements
        for lib in get_libraries(chip, 'logic'):
            # mandatory for logiclibs
            chip.add('tool', tool, 'task', task, 'require',
                     ",".join(_get_synthesis_library_key(chip, lib, syn_corners)),
                     step=step, index=index)

        for lib in get_libraries(chip, 'macro'):
            # optional for macrolibs
            if chip.valid(*_get_synthesis_library_key(chip, lib, syn_corners)):
                chip.add('tool', tool, 'task', task, 'require',
                         ",".join(_get_synthesis_library_key(chip, lib, syn_corners)),
                         step=step, index=index)
            elif chip.valid('library', lib, 'output', 'blackbox', 'verilog'):
                chip.add('tool', tool, 'task', task, 'require',
                         ",".join(['library', lib, 'output', 'blackbox', 'verilog']),
                         step=step, index=index)

    if chip.valid('input', 'constraint', 'sdc') and \
       chip.get('input', 'constraint', 'sdc', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,constraint,sdc',
                 step=step, index=index)

    # set default control knobs
    mainlib = get_mainlib(chip)
    for option, value in [
            ('flatten', "true"),
            ('auto_flatten', "true"),
            ('hier_iterations', "10"),
            ('hier_threshold', "1000"),
            ('autoname', "true"),
            ('add_buffers', "true")]:
        chip.set('tool', tool, 'task', task, 'var', option, value,
                 step=step, index=index, clobber=False)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', option]),
                 step=step, index=index)

    # Add conditionally required mainlib variables
    if chip.valid('library', mainlib, 'option', 'var', 'yosys_buffer_cell'):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'var', 'yosys_buffer_input']),
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'var', 'yosys_buffer_output']),
                 step=step, index=index)

    library_has_addermap = \
        chip.valid('library', mainlib, 'option', 'file', 'yosys_addermap') and \
        chip.get('library', mainlib, 'option', 'file', 'yosys_addermap')
    chip.set('tool', tool, 'task', task, 'var', 'map_adders',
             "true" if library_has_addermap else "false",
             step=step, index=index, clobber=False)
    if chip.get('tool', tool, 'task', task, 'var', 'map_adders', step=step, index=index)[0] == \
       "true":
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'map_adders']),
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'file', 'yosys_addermap']),
                 step=step, index=index)

    for var0, var1 in [('memory_libmap', 'memory_techmap')]:
        key0 = ['tool', tool, 'tak', task, 'file', var0]
        key1 = ['tool', tool, 'tak', task, 'file', var1]
        if chip.valid(*key0) and chip.get(*key0, step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require', ",".join(key1), step=step, index=index)
        if chip.valid(*key1) and chip.get(*key1, step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require', ",".join(key0), step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'synthesis_corner', get_synthesis_corner(chip),
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'synthesis_corner']),
             step=step, index=index)

    # Require abc clock conversion factor, from library units to ps
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['library', mainlib, 'option', 'var', 'yosys_abc_clock_multiplier']),
             step=step, index=index)
    abc_driver = get_abc_driver(chip)
    if abc_driver:
        chip.set('tool', tool, 'task', task, 'var', 'abc_constraint_driver', abc_driver,
                 step=step, index=index, clobber=False)

    set_tool_task_var(chip, 'abc_constraint_load',
                      schelp='Capacitive load for the abc techmapping in fF, '
                             'if not specified it will not be used',
                      require='lib')

    # document parameters
    chip.set('tool', tool, 'task', task, 'var', 'preserve_modules',
             'List of modules in input files to prevent flatten from "flattening"', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'blackbox_modules',
             'List of modules in input files to exclude from synthesis by replacing them '
             'with empty blackboxes"', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'flatten',
             'true/false, invoke synth with the -flatten option', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'autoname',
             'true/false, call autoname to rename wires based on registers', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'map_adders',
             'true/false, techmap adders in Yosys', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'synthesis_corner',
             'Timing corner to use for synthesis', field='help')
    chip.set('tool', tool, 'task', task, 'file', 'dff_liberty',
             'Liberty file to use for flip-flop mapping, if not specified the first in the '
             'logiclib is used', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'abc_constraint_driver',
             'Buffer that drives the abc techmapping, defaults to first buffer specified',
             field='help')
    chip.set('tool', tool, 'task', task, 'file', 'abc_constraint_file',
             'File used to pass in constraints to abc', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'abc_clock_period',
             'Clock period to use for synthesis in ps, if more than one clock is specified, the '
             'smallest period is used.', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'abc_clock_derating',
             'Used to derate the clock period to further constrain the clock, '
             'values between 0 and 1', field='help')
    chip.set('tool', tool, 'task', task, 'file', 'techmap',
             'File to use for techmapping in Yosys', field='help')
    chip.set('tool', tool, 'task', task, 'file', 'dff_liberty_file',
             'File to use for the DFF mapping stage of Yosys', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'add_buffers',
             'true/false, flag to indicate whether to add buffers or not.', field='help')

    chip.set('tool', tool, 'task', task, 'var', 'auto_flatten',
             'true/false, attempt to determine how to flatten the design', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'hier_iterations',
             'Number of iterations to attempt to determine the hierarchy to flatten',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'hier_threshold',
             'Instance limit for the number of cells in a module to preserve.',
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'strategy',
             'ABC synthesis strategy. Allowed values are DELAY0-4, AREA0-3, or if the strategy '
             'starts with a + it is assumed to be actual commands for ABC.',
             field='help')

    chip.set('tool', tool, 'task', task, 'file', 'memory_libmap',
             'File used to map memories with yosys', field='help')
    chip.set('tool', tool, 'task', task, 'file', 'memory_techmap',
             'File used to techmap memories with yosys', field='help')

    chip.add('tool', tool, 'task', task, 'file', 'synth_extra_map',
             'tools/yosys/techmaps/lcu_kogge_stone.v', package='siliconcompiler',
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'file', 'synth_extra_map',
             'Files used in synthesis to perform additional techmapping', field='help')


################################
# mark cells dont use and format liberty files for yosys and abc
################################
def prepare_synthesis_libraries(chip):
    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    corners = chip.get('tool', tool, 'task', task, 'var', 'synthesis_corner',
                       step=step, index=index)

    # mark dff libery file with dont use
    dff_liberty_file = chip.find_files('tool', tool, 'task', task, 'file', 'dff_liberty',
                                       step=step, index=index)[0]
    yosys_dff_file = chip.get('tool', tool, 'task', task, 'file', 'dff_liberty_file',
                              step=step, index=index)[0]

    with open(yosys_dff_file, 'w') as f:
        f.write(prepareLib.processLibertyFile(
            dff_liberty_file,
            logger=None if chip.get('option', 'quiet', step=step, index=index) else chip.logger
        ))

    # Clear in case of rerun
    for libtype in ('synthesis_libraries', 'synthesis_libraries_macros'):
        chip.set('tool', tool, 'task', task, 'file', libtype, [],
                 step=step, index=index)

    # Generate synthesis_libraries and synthesis_macro_libraries for Yosys use

    # mark libs with dont_use since ABC cannot get this information via its commands
    # this also ensures the liberty files have been decompressed and corrected formatting
    # issues that generally cannot be handled by yosys or yosys-abc
    def get_synthesis_libraries(lib):
        keypath = _get_synthesis_library_key(chip, lib, corners)
        if chip.valid(*keypath):
            return chip.find_files(*keypath, step=step, index=index)
        return []

    for libtype in ('logic', 'macro'):
        for lib in get_libraries(chip, libtype):
            lib_content = {}
            # Mark dont use
            for lib_file in get_synthesis_libraries(lib):
                # Ensure a unique name is used for library
                lib_file_name_base = os.path.basename(lib_file)
                if lib_file_name_base.lower().endswith('.gz'):
                    lib_file_name_base = lib_file_name_base[0:-3]
                if lib_file_name_base.lower().endswith('.lib'):
                    lib_file_name_base = lib_file_name_base[0:-4]

                lib_file_name = lib_file_name_base
                unique_ident = 0
                while lib_file_name in lib_content:
                    lib_file_name = f'{lib_file_name_base}_{unique_ident}'
                    unique_ident += 1

                if lib_file == dff_liberty_file:
                    with sc_open(yosys_dff_file) as f:
                        lib_content[lib_file_name] = f.read()
                    continue

                lib_content[lib_file_name] = prepareLib.processLibertyFile(
                        lib_file,
                        logger=None if chip.get('option', 'quiet',
                                                step=step, index=index) else chip.logger)

            if not lib_content:
                continue

            var_name = 'synthesis_libraries'
            if libtype == "macro":
                var_name = 'synthesis_libraries_macros'

            for file, content in lib_content.items():
                output_file = os.path.join(
                    chip.getworkdir(step=step, index=index),
                    'inputs',
                    f'sc_{libtype}_{lib}_{file}.lib'
                )

                with open(output_file, 'w') as f:
                    f.write(content)

                chip.add('tool', tool, 'task', task, 'file', var_name, output_file,
                         step=step, index=index)


def create_abc_synthesis_constraints(chip):

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    abc_driver = chip.get('tool', tool, 'task', task, 'var', 'abc_constraint_driver',
                          step=step, index=index)
    if abc_driver:
        abc_driver = abc_driver[0]

    abc_load = chip.get('tool', tool, 'task', task, 'var', 'abc_constraint_load',
                        step=step, index=index)
    if abc_load:
        abc_load = abc_load[0]

    if not abc_driver and not abc_load:
        # neither is set so nothing to do
        return

    with open(chip.get('tool', tool, 'task', task, 'file', 'abc_constraint_file',
                       step=step, index=index)[0], "w") as f:
        if abc_load:
            # convert to fF
            abc_load = re.match(r'([0-9]*\.?[0-9]*)\s*([fpa])[fF]?', abc_load)

            if not abc_load:
                raise ValueError(f'Unable to parse {abc_load}')

            abc_load_unit = abc_load.group(2)
            if not abc_load_unit:
                abc_load_unit = chip.get('unit', 'capacitance')[0]

            abc_load = float(abc_load.group(1))

            if abc_load_unit == 'p':
                abc_load *= 1000
            elif abc_load_unit == 'a':
                abc_load /= 1000

        abc_template = utils.get_file_template('abc.const',
                                               root=os.path.join(os.path.dirname(__file__),
                                                                 'templates'))
        f.write(abc_template.render(abc_driver=abc_driver, abc_load=abc_load))


def get_synthesis_corner(chip):
    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    syn_corners = chip.get('tool', tool, 'task', task, 'var', 'synthesis_corner',
                           step=step, index=index)
    if syn_corners:
        return syn_corners

    # determine corner based on setup corner from constraints
    corner = None
    for constraint in chip.getkeys('constraint', 'timing'):
        checks = chip.get('constraint', 'timing', constraint, 'check', step=step, index=index)
        if "setup" in checks and not corner:
            corner = chip.get('constraint', 'timing', constraint, 'libcorner',
                              step=step, index=index)

    if not corner:
        # try getting it from first constraint with a valid libcorner
        for constraint in chip.getkeys('constraint', 'timing'):
            if not corner:
                corner = chip.get('constraint', 'timing', constraint, 'libcorner',
                                  step=step, index=index)

    return corner


def _get_synthesis_library_key(chip, lib, corners):
    if chip.valid('library', lib, 'option', 'file', 'yosys_synthesis_libraries'):
        return ('library', lib, 'option', 'file', 'yosys_synthesis_libraries')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    delaymodel = chip.get('asic', 'delaymodel', step=step, index=index)
    for corner in corners:
        if chip.valid('library', lib, 'output', corner, delaymodel):
            return ('library', lib, 'output', corner, delaymodel)

    return ('library', lib, 'output', corners[0], delaymodel)


def get_dff_liberty_file(chip):
    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    dff_liberty = None
    if chip.valid('tool', tool, 'task', task, 'file', 'dff_liberty'):
        dff_liberty = chip.find_files('tool', tool, 'task', task, 'file', 'dff_liberty',
                                      step=step, index=index)
        if dff_liberty:
            return dff_liberty[0]

    mainlib = get_mainlib(chip)
    if chip.valid('library', mainlib, 'option', 'file', 'yosys_dff_liberty'):
        dff_liberty = chip.find_files('library', mainlib, 'option', 'file', 'yosys_dff_liberty')
        if dff_liberty:
            return dff_liberty[0]

    corners = get_synthesis_corner(chip)
    if corners is None:
        return None

    # if dff liberty file is not set, use the first liberty file defined
    for lib in get_libraries(chip, 'logic'):
        if not chip.valid(*_get_synthesis_library_key(chip, lib, corners)):
            continue

        lib_files = chip.find_files(*_get_synthesis_library_key(chip, lib, corners),
                                    step=step, index=index)
        if len(lib_files) > 0:
            return lib_files[0]

    return None


def get_abc_period(chip):

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    mainlib = get_mainlib(chip)

    abc_clock_period = chip.get('tool', tool, 'task', task, 'var', 'abc_clock_period',
                                step=step, index=index)
    if abc_clock_period:
        return abc_clock_period[0]

    period = None

    abc_clock_multiplier = float(chip.get('library', mainlib, 'option', 'var',
                                          'yosys_abc_clock_multiplier')[0])

    # get clock information from sdc files
    # TODO: fix for fpga/asic differentiation later
    if chip.valid('input', 'constraint', 'sdc'):
        for sdc in chip.find_files('input', 'constraint', 'sdc', step=step, index=index):
            lines = []
            with sc_open(sdc) as f:
                lines = f.read().splitlines()

            # collect simple variables in case clock is specified with a variable
            re_var = r"[A-Za-z0-9_]+"
            re_num = r"[0-9\.]+"
            sdc_vars = {}
            for line in lines:
                tcl_variable = re.findall(fr"^\s*set\s+({re_var})\s+({re_num})", line)
                if tcl_variable:
                    var_name, var_value = tcl_variable[0]
                    sdc_vars[f'${var_name}'] = float(var_value)

            # TODO: handle line continuations
            for line in lines:
                clock_period = re.findall(fr"create_clock\s.*-period\s+({re_num}|\${re_var})",
                                          line)
                if clock_period:
                    clock_period = clock_period[0]
                    if clock_period[0] == '$':
                        if clock_period in sdc_vars:
                            clock_period = sdc_vars[clock_period]
                        else:
                            chip.logger.warning('Unable to identify clock period from '
                                                f'{clock_period}.')
                            continue
                    else:
                        clock_period = float(clock_period)

                    clock_period = clock_period * abc_clock_multiplier

                    if period is None:
                        period = clock_period
                    else:
                        period = min(period, clock_period)

    if period is None:
        # get clock information from defined clocks
        for pin in chip.getkeys('datasheet', 'pin'):
            for mode in chip.getkeys('datasheet', 'pin', pin, 'type'):
                if chip.get('datasheet', 'pin', pin, 'type', mode) == 'clock':
                    clock_period = min(chip.get('datasheet', 'pin', pin, 'tperiod', mode)) * 1e12

                    if period is None:
                        period = clock_period
                    else:
                        period = min(period, clock_period)

    if period is None:
        return None

    abc_clock_derating = chip.get('tool', tool, 'task', task, 'var', 'abc_clock_derating',
                                  step=step, index=index)
    if abc_clock_derating:
        derating = float(abc_clock_derating[0])
        if derating > 1:
            chip.logger.warning("abc_clock_derating is greater than 1.0")
        elif derating > 0:
            period *= (1.0 - derating)
        else:
            chip.logger.error("abc_clock_derating is negative")

    return int(period)


def get_abc_driver(chip):

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    abc_driver = chip.get('tool', tool, 'task', task, 'var', 'abc_constraint_driver',
                          step=step, index=index)
    if abc_driver:
        return abc_driver[0]

    abc_driver = None
    # get the first driver defined in the logic lib
    for lib in get_libraries(chip, 'logic'):
        if chip.valid('library', lib, 'option', 'var', 'yosys_driver_cell') and not abc_driver:
            abc_driver = chip.get('library', lib, 'option', 'var', 'yosys_driver_cell')[0]

    return abc_driver


##################################################
def pre_process(chip):
    ''' Tool specific function to run before step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # copy techmapping from libraries
    logiclibs = get_libraries(chip, 'logic')
    macrolibs = get_libraries(chip, 'macro')
    for lib in logiclibs + macrolibs:
        if not chip.valid('library', lib, 'option', 'file', 'yosys_techmap'):
            continue
        for techmap in chip.find_files('library', lib, 'option', 'file', 'yosys_techmap'):
            if techmap is None:
                continue
            chip.add('tool', tool, 'task', task, 'file', 'techmap', techmap, step=step, index=index)

    # Constants needed by yosys, do not allow overriding of values so force clobbering
    chip.set('tool', tool, 'task', task, 'file', 'dff_liberty_file',
             f"{chip.getworkdir(step=step, index=index)}/inputs/sc_dff_library.lib",
             step=step, index=index, clobber=True)
    chip.set('tool', tool, 'task', task, 'file', 'abc_constraint_file',
             f"{chip.getworkdir(step=step, index=index)}/inputs/sc_abc.constraints",
             step=step, index=index, clobber=True)

    dff_liberty_file = get_dff_liberty_file(chip)
    if dff_liberty_file:
        chip.set('tool', tool, 'task', task, 'file', 'dff_liberty', dff_liberty_file,
                 step=step, index=index, clobber=False)

    abc_clock_period = get_abc_period(chip)
    if abc_clock_period:
        chip.set('tool', tool, 'task', task, 'var', 'abc_clock_period', str(abc_clock_period),
                 step=step, index=index, clobber=False)

    prepare_synthesis_libraries(chip)
    create_abc_synthesis_constraints(chip)
    return


def post_process(chip):
    syn_post_process(chip)



#!/usr/bin/env python3

import argparse
import os
import sys
import tarfile
from siliconcompiler import Chip, Schema
from siliconcompiler.package import _path as sc_path
from siliconcompiler.scheduler import _runtask, _executenode
from siliconcompiler import __version__


##########################
def main():
    schema = Schema()

    # Can't use chip.cmdline because we don't want a bunch of extra logger information
    parser = argparse.ArgumentParser(prog='run_node',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description='Script to run a single node in an SC flowgraph')

    parser.add_argument('-version',
                        action='version',
                        version=__version__)
    parser.add_argument('-cfg',
                        required=True,
                        metavar='<file>',
                        help=schema.get('option', 'cfg',
                                        field='shorthelp'))
    parser.add_argument('-cwd',
                        required=True,
                        metavar='<directory>',
                        help='Run current working directory')
    parser.add_argument('-builddir',
                        metavar='<directory>',
                        required=True,
                        help=schema.get('option', 'builddir',
                                        field='shorthelp'))
    parser.add_argument('-cachedir',
                        metavar='<directory>',
                        required=True,
                        help=schema.get('option', 'cachedir',
                                        field='shorthelp'))
    parser.add_argument('-cachemap',
                        metavar='<package>:<directory>',
                        nargs='+',
                        help='Map of caches to prepopulate runner with')
    parser.add_argument('-step',
                        required=True,
                        metavar='<step>',
                        help=schema.get('arg', 'step',
                                        field='shorthelp'))
    parser.add_argument('-index',
                        required=True,
                        metavar='<index>',
                        help=schema.get('arg', 'index',
                                        field='shorthelp'))
    parser.add_argument('-remoteid',
                        metavar='<id>',
                        help=schema.get('record', 'remoteid',
                                        field='shorthelp'))
    parser.add_argument('-archive',
                        metavar='<file>',
                        help='Generate archive')
    parser.add_argument('-include',
                        metavar='<path>',
                        nargs='+',
                        help='Files to include in archive')
    parser.add_argument('-unset_scheduler',
                        action='store_true',
                        help='Unset scheduler to ensure local run')
    args = parser.parse_args()

    # Change to working directory to allow rel path to be build dir
    # this avoids needing to deal with the job hash on the client
    # side
    os.chdir(args.cwd)

    # Create the Chip object.
    chip = Chip('<design>')
    chip.read_manifest(args.cfg)

    # setup work directory
    chip.set('arg', 'step', args.step)
    chip.set('arg', 'index', args.index)
    chip.set('option', 'builddir', args.builddir)
    chip.set('option', 'cachedir', args.cachedir)

    if args.remoteid:
        chip.set('record', 'remoteid', args.remoteid)

    if args.unset_scheduler:
        for vals, step, index in chip.schema._getvals('option', 'scheduler', 'name'):
            chip.unset('option', 'scheduler', 'name', step=step, index=index)

    # Init logger to ensure consistent view
    chip._init_logger(step=chip.get('arg', 'step'),
                      index=chip.get('arg', 'index'),
                      in_run=True)

    if args.cachemap:
        for cachepair in args.cachemap:
            package, path = cachepair.split(':')
            chip._packages[package] = path

    # Populate cache without downloading
    for package in chip.getkeys('package', 'source'):
        sc_path(chip, package, None)

    # Run the task.
    error = True
    try:
        _runtask(chip,
                 chip.get('option', 'flow'),
                 chip.get('arg', 'step'),
                 chip.get('arg', 'index'),
                 _executenode)
        error = False

    finally:
        if args.archive:
            # Archive the results.
            with tarfile.open(args.archive,
                              mode='w:gz') as tf:
                chip._archive_node(tf,
                                   step=args.step,
                                   index=args.index,
                                   include=args.include)

    # Return success/fail flag, in case the caller is interested.
    if error:
        return 1
    return 0


##########################
if __name__ == "__main__":
    sys.exit(main())



import siliconcompiler
from lambdapdk.asap7.libs.asap7sc7p5t import setup


#########################
if __name__ == "__main__":
    for lib in setup(siliconcompiler.Chip('<lib>')):
        lib.write_manifest(f'{lib.top()}.json')



import os

from siliconcompiler.tools._common import input_provides, get_tool_task
from siliconcompiler.tools._common.asic import set_tool_task_var
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.openroad import post_process as or_post_process


def setup(chip):
    '''
    Perform floorplanning, pin placements, macro placements and power grid generation
    '''

    # Generic tool setup.
    # default tool settings, note, not additive!

    tool = 'openroad'
    script = 'sc_rdlroute.tcl'
    refdir = os.path.join('tools', tool, 'scripts')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    chip.set('tool', tool, 'exe', tool)
    chip.set('tool', tool, 'vswitch', '-version')
    chip.set('tool', tool, 'version', '>=v2.0-16839')
    chip.set('tool', tool, 'format', 'tcl')

    # exit automatically in batch mode and not breakpoint
    option = ''
    if exit and not chip.get('option', 'breakpoint', step=step, index=index):
        option += " -exit"

    option += " -metrics reports/metrics.json"
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index)

    # Input/Output requirements for default asicflow steps

    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler')
    chip.set('tool', tool, 'task', task, 'script', script,
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    if chip.get('option', 'nodisplay'):
        # Tells QT to use the offscreen platform if nodisplay is used
        chip.set('tool', tool, 'task', task, 'env', 'QT_QPA_PLATFORM', 'offscreen',
                 step=step, index=index)

    # basic warning and error grep check on logfile
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', r'^\[WARNING|^Warning',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', r'^\[ERROR',
             step=step, index=index, clobber=False)

    chip.add('tool', tool, 'task', task, 'require',
             'option,var,openroad_libtype',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'file', 'rdlroute']),
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'file', 'rdlroute',
             'script to perform rdl route',
             field='help')

    set_tool_task_var(chip, param_key='fin_add_fill',
                      default_value='false',
                      schelp='true/false, when true enables adding fill, '
                             'if enabled by the PDK, to the design',
                      skip='lib')

    if f'{design}.v' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.v', step=step, index=index)
    elif f'{design}.vg' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.vg', step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['input', 'netlist', 'verilog']),
                 step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', design + '.sdc', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.vg', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.def', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.odb', step=step, index=index)


def pre_process(chip):
    build_pex_corners(chip)


def post_process(chip):
    or_post_process(chip)



# Copyright 2024 Silicon Compiler Authors. All Rights Reserved.

import argparse
import glob
import subprocess
import sys
import shutil
import re
import os.path
from collections.abc import Container
from pathlib import Path
import siliconcompiler
from siliconcompiler.scheduler import _get_machine_info


class ChoiceOptional(Container):
    def __init__(self, choices):
        super().__init__()

        self.__choices = set(choices)

    def __contains__(self, item):
        if not item:
            # allow empty value
            return True
        return item in self.__choices

    def __iter__(self):
        return self.__choices.__iter__()

    def get_items(self, choices):
        items = set(choices)
        return sorted(list(items))


def install_tool(tool, script, build_dir, prefix):
    # Ensure build dir is available
    build_dir = Path(build_dir) / tool
    shutil.rmtree(str(build_dir), ignore_errors=True)
    build_dir.mkdir(parents=True, exist_ok=True)

    # setup environment
    env = os.environ.copy()
    env["PREFIX"] = prefix

    # run
    ret = subprocess.call(script, env=env, cwd=build_dir)
    if ret != 0:
        print(f"Error occurred while building/installing {tool}")
        return False
    return True


def show_tool(tool, script):
    def print_header(head):
        border_len = max(80, len(script) + 2)
        border = border_len*"#"
        print(border)
        print(f"# {tool} script / {head}")
        if head == "start":
            print(f"# {script}")
        print(border)

    print_header("start")

    with open(script) as f:
        for line in f:
            print(line.rstrip())

    print_header("end")


def _get_os_name():
    machine_info = _get_machine_info()
    system = machine_info.get('system', "").lower()
    distro = machine_info.get('distro', "").lower()
    osversion = machine_info.get('osversion', "").lower()
    if system == 'linux':
        if distro == 'ubuntu':
            version, _ = osversion.split('.')
            return f"{distro}{version}"
        elif distro == 'rocky':
            version, _ = osversion.split('.')
            return f"rhel{version}"
        elif distro == 'rhel':
            version, _ = osversion.split('.')
            return f"rhel{version}"
    return None


def print_machine_info():
    machine_info = _get_machine_info()
    mapped_os = _get_os_name()

    print("System:   ", machine_info.get('system', None))
    print("Distro:   ", machine_info.get('distro', None))
    print("Version:  ", machine_info.get('osversion', None))
    print("Mapped OS:", mapped_os)
    print("Scripts:  ", _get_tool_script_dir())


def _get_tool_script_dir():
    return Path(siliconcompiler.__file__).parent / "toolscripts"


def _get_tools_list():
    tools_root = _get_tool_script_dir()

    script_dir = None
    os_dir = _get_os_name()
    if os_dir:
        script_dir = tools_root / os_dir
        if not script_dir.exists():
            script_dir = None

    tools = {}
    if script_dir:
        for script in glob.glob(str(script_dir / "install-*.sh")):
            tool = re.match(r"install-(.*)\.sh", os.path.basename(script).lower())
            tools[tool.group(1)] = script

    return tools


def _recommended_tool_groups(tools):
    groups = {
        "asic": {"surelog", "sv2v", "yosys", "openroad", "klayout"},
        "fpga": {"surelog", "sv2v", "yosys", "vpr"},
        "digital-simulation": {"verilator", "icarus"},
        "analog-simulation": {"xyce"}
    }

    filter_groups = {}
    for group, group_tools in groups.items():
        if all([tool in tools for tool in group_tools]):
            filter_groups[group] = group_tools
    return filter_groups


def main():
    progname = "sc-install"
    description = """
-----------------------------------------------------------
SC app install supported tools.

To install a single tool:
    sc-install openroad

To install multiple tools:
    sc-install openroad yosys klayout

To install a group of tools:
    sc-install -group asic

To install tools in a different location:
    sc-install -prefix /usr/local yosys

To build tools in a different location:
    sc-install -build_dir /tmp yosys

To show the install script:
    sc-install -show openroad

To system debugging information (this should only be used to debug):
    sc-install -debug_machine
-----------------------------------------------------------
"""
    parser = argparse.ArgumentParser(
        prog=progname,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    tools = _get_tools_list()

    tool_choices = ChoiceOptional(tools.keys())
    parser.add_argument(
        "tool",
        nargs="*",
        choices=tool_choices,
        help="tool to install")

    tool_groups = _recommended_tool_groups(tools)
    parser.add_argument(
        "-group",
        nargs="+",
        choices=tool_groups.keys(),
        help=f"tool group to install{' - not supported' if not tool_groups else ''}")

    parser.add_argument(
        "-prefix",
        default=Path.home() / ".local",
        help="Prefix to use when installing tool",
        metavar="<path>")

    parser.add_argument(
        "-build_dir",
        default=Path.home() / ".sc" / "tool_build",
        help="Directory to build the tool in",
        metavar="<path>")

    parser.add_argument(
        "-show",
        action="store_true",
        help="Show the install script and exit")

    parser.add_argument(
        "-debug_machine",
        action="store_true",
        help="Show information about this machine and exit")

    args = parser.parse_args()

    if args.debug_machine:
        print_machine_info()
        return 0

    if not args.tool:
        args.tool = []

    args.tool = list(args.tool)
    if args.group:
        for group in args.group:
            args.tool.extend(tool_groups[group])

    tools_handled = set()
    for tool in args.tool:
        if tool in tools_handled:
            continue
        tools_handled.add(tool)
        if args.show:
            show_tool(tool, tools[tool])
        else:
            if not install_tool(tool, tools[tool], args.build_dir, args.prefix):
                return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



import os
from siliconcompiler.tools._common import input_provides


def setup(chip):
    '''
    Convert SystemVerilog to verilog
    '''

    chip.logger.debug("Setting up sv2v")

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    topmodule = chip.top()
    if topmodule + ".sv" not in input_provides(chip, step, index):
        return "will not receive systemverilog to convert"

    tool = 'sv2v'
    task = 'convert'

    chip.set('tool', tool, 'exe', tool)
    chip.set('tool', tool, 'vswitch', '--numeric-version')
    chip.set('tool', tool, 'version', '>=0.0.9', clobber=False)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    # Since we run sv2v after the import/preprocess step, there should be no
    # need for specifying include dirs/defines. However we don't want to pass
    # --skip-preprocessor because there may still be unused preprocessor
    # directives not removed by the importer and passing the --skip-preprocessor
    # flag would cause sv2v to error.

    # since this step should run after import, the top design module should be
    # set and we can read the pickled Verilog without accessing the original
    # sources
    topmodule = chip.top()
    chip.set('tool', tool, 'task', task, 'option', [], step=step, index=index)
    chip.add('tool', tool, 'task', task, 'option', "inputs/" + topmodule + ".sv",
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'option', "--write=outputs/" + topmodule + ".v",
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'input', f'{topmodule}.sv', step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', f'{topmodule}.v', step=step, index=index)



from siliconcompiler.tools.builtin import _common
import re
from siliconcompiler.tools.builtin.builtin import set_io_files
from siliconcompiler import flowgraph, SiliconCompilerError


def setup(chip):
    '''
    Selects a task from a list of inputs.

    The selector criteria provided is used to create a custom function
    for selecting the best step/index pair from the inputs. Metrics and
    weights are passed in and used to select the step/index based on
    the minimum or maximum score depending on the 'op' argument from
    ['flowgraph', flow, step, index, 'args'] in the form 'minimum(metric)' or
    'maximum(metric)'.

    The function can be used to bypass the flows weight functions for
    the purpose of conditional flow execution and verification.
    '''

    set_io_files(chip)


def _select_inputs(chip, step, index):
    chip.logger.info("Running builtin task 'mux'")

    flow = chip.get('option', 'flow')
    inputs = chip.get('flowgraph', flow, step, index, 'input')
    arguments = chip.get('flowgraph', flow, step, index, 'args')

    operations = []
    for criteria in arguments:
        m = re.match(r'(minimum|maximum)\((\w+)\)', criteria)
        if not m:
            raise SiliconCompilerError(f"Illegal mux criteria: {criteria}", chip=chip)

        op = m.group(1)
        metric = m.group(2)
        if metric not in chip.getkeys('metric'):
            raise SiliconCompilerError(
                f"Criteria must use legal metrics only: {criteria}", chip=chip)

        operations.append((metric, op))
    score, sel_inputs = _common._mux(chip, *inputs, operations=operations)

    if sel_inputs:
        chip.logger.info(f"Selected '{sel_inputs[0]}{sel_inputs[1]}' with score {score:.3f}")

    return sel_inputs


def _gather_outputs(chip, step, index):
    flow = chip.get('option', 'flow')

    in_nodes = chip.get('flowgraph', flow, step, index, 'input')
    in_task_outputs = [flowgraph._gather_outputs(chip, *node) for node in in_nodes]

    if len(in_task_outputs) > 0:
        return in_task_outputs[0].intersection(*in_task_outputs[1:])

    return []


def run(chip):
    return _common.run(chip)


def post_process(chip):
    _common.post_process(chip)



from siliconcompiler import SiliconCompilerError
from siliconcompiler.tools.klayout.klayout import setup as setup_tool
from siliconcompiler.tools._common import input_provides, get_tool_task


def setup(chip):
    '''
    Perform unit operations on stream files. Currently supports:

        * rotating (rotate)
        * renaming (rename)
        * merging streams (merge)
        * adding streams together (add)
        * adding outline to top (outline)
        * swapping cells (swap)
        * adding new top cell (add_top)
        * renaming cells (rename_cell)
        * flatten
        * deleting layers
        * merging shapes
        * writing (write)
        * converting properties into text labels on design (convert_property)

    To rotate:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', 'rotate')

    To rename:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'rename:tool,klayout,task,operations,var,new_name')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'var', 'new_name', \\
        'chip_top')

    To merge streams:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'merge:tool,klayout,task,operations,file,fill_stream')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'file', 'fill_stream', \\
        './fill.gds')

    or to get it from the inputs to this task:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'merge:fill.gds')

    To add streams:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'add:tool,klayout,task,operations,file,fill_stream')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'file', 'fill_stream', \\
        './fill.gds')

    or to get it from the inputs to this task:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'add:fill.gds')

    To add outline:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'outline:tool,klayout,task,operations,var,outline')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'var', 'outline', \\
        ['10', '1'])  # layer / purpose pair

    To swap layout cells:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'swap:tool,klayout,task,operations,var,cell_swap')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'var', 'cell_swap', \\
        ['dummy_ANDX2=ANDX2', 'dummy_NANDX2=NANDX2'])

    To rename cells:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'rename_cell:tool,klayout,task,operations,var,rename_cell')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'var', 'rename_cell', \\
        ['dummy_ANDX2=ANDX2', 'dummy_NANDX2=NANDX2'])

    To add new top cell:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'add_top:tool,klayout,task,operations,var,new_name')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'var', 'new_name', \\
        'chip_top')

    To write out a new file:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'write:combined.gds')

    To convert stream properties to text labels:

    >>> chip.add('tool', 'klayout, 'task', 'operations', 'var', 'operations', \\
        'convert_property:tool,klayout,task,operations,var,convert_c4_bumps')
    >>> chip.set('tool', 'klayout, 'task', 'operations', 'var', 'convert_c4_bumps', \\
        ['10', '2', \\  # layer / purpose pair for the source of the labels
         '3' \\  # stream property number
         '85', '5'])  #  (optional) destination layer / purpose pair, if not provided
                      # the source pair will be used instead.
    '''

    # Generic tool setup.
    setup_tool(chip)

    tool = 'klayout'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    clobber = False

    chip.set('tool', tool, 'task', task, 'threads', 1, step=step, index=index, clobber=clobber)

    script = 'klayout_operations.py'
    option = ['-z', '-nc', '-rx', '-r']
    chip.set('tool', tool, 'task', task, 'script', script, step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index, clobber=clobber)

    # Set stream extension
    streams = ('gds', 'oas')
    chip.set('tool', tool, 'task', task, 'var', 'stream', 'gds',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'stream',
             f'Extension to use for stream generation ({streams})',
             field='help')
    default_stream = chip.get('tool', tool, 'task', task, 'var', 'stream',
                              step=step, index=index)[0]
    # Input/Output requirements for default flow
    design = chip.top()
    if f'{design}.{default_stream}' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.{default_stream}',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', f'input,layout,{default_stream}')
    chip.add('tool', tool, 'task', task, 'output', f'{design}.{default_stream}',
             step=step, index=index)

    # Export GDS with timestamps by default.
    chip.set('tool', tool, 'task', task, 'var', 'timestamps', 'true',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'timestamps',
             'Export GDSII with timestamps',
             field='help')

    klayout_ops = ('merge',
                   'add',
                   'rotate',
                   'outline',
                   'rename',
                   'swap',
                   'add_top',
                   'write',
                   'convert_property',
                   'flatten',
                   'merge_shapes',
                   'delete_layers',
                   'rename_cell')
    ops = chip.get('tool', tool, 'task', task, 'var', 'operations', step=step, index=index)
    for op in ops:
        klayout_op = op.split(':', 1)
        if len(klayout_op) == 1:
            klayout_op = klayout_op[0]
            args = ""
        else:
            klayout_op, args = klayout_op

        if klayout_op not in klayout_ops:
            raise SiliconCompilerError(
                f'{klayout_op} is not a supported operation in klayout: {klayout_ops}',
                chip=chip)

        if klayout_op in ('add', 'merge'):
            if ',' in args:
                chip.add('tool', tool, 'task', task, 'require', args, step=step, index=index)
            elif args:
                chip.add('tool', tool, 'task', task, 'input', args, step=step, index=index)
            else:
                raise SiliconCompilerError(
                    f'{klayout_op} requires a filename to read or a keypath', chip=chip)
        elif klayout_op in ('outline',
                            'rename',
                            'swap',
                            'add_top',
                            'convert_property',
                            'merge_shapes',
                            'delete_layers',
                            'rename_cell'):
            chip.add('tool', tool, 'task', task, 'require', args, step=step, index=index)
        elif klayout_op in ('rotate', 'flatten'):
            if args:
                raise SiliconCompilerError('rotate does not take any arguments', chip=chip)
        elif klayout_op in ('write'):
            if not args:
                raise SiliconCompilerError('write requires a filename to save to', chip=chip)
            chip.add('tool', tool, 'task', task, 'output', args,
                     step=step, index=index)



from siliconcompiler.schema.schema_cfg import scparam
from siliconcompiler.schema import Schema


SCHEMA_VERSION = '0.0.2'


def schema_cfg():
    # Basic schema setup
    cfg = {}

    scparam(cfg, ['schemaversion'],
            sctype='str',
            scope='global',
            defvalue=SCHEMA_VERSION,
            require='all',
            shorthelp="Schema version number",
            lock=True,
            switch="-schemaversion <str>",
            example=["api: server.get('schemaversion')"],
            schelp="""SiliconCompiler server schema version number.""")

    scparam(cfg, ['option', 'port'],
            sctype='int',
            scope='global',
            defvalue=8080,
            require='all',
            shorthelp="Port number to run the server on.",
            switch="-port <int>",
            example=["cli: -port 8000",
                     "api: server.set('option', 'port', 8080)"],
            schelp="""Port number to run the server on.""")

    scparam(cfg, ['option', 'cluster'],
            sctype='enum',
            enum=['local', 'slurm'],
            scope='global',
            defvalue='local',
            require='all',
            shorthelp="Type of compute cluster to use.",
            switch="-cluster <str>",
            example=["cli: -cluster slurm",
                     "api: server.set('option', 'clister', 'slurm')"],
            schelp="""Type of compute cluster to use.""")

    scparam(cfg, ['option', 'nfsmount'],
            sctype='dir',
            scope='global',
            defvalue='/nfs/sc_compute',
            require='all',
            shorthelp="Directory of mounted shared NFS storage.",
            switch="-nfsmount <dir>",
            example=["cli: -nfsmount ~/sc_server",
                     "api: server.set('option', 'server', '~/sc_server')"],
            schelp="""Directory of mounted shared NFS storage.""")

    scparam(cfg, ['option', 'auth'],
            sctype='bool',
            scope='global',
            defvalue=False,
            require='all',
            shorthelp="Flag determining whether to enable authenticated and encrypted jobs.",
            switch="-auth <bool>",
            example=["cli: -auth true",
                     "api: server.set('option', 'auth', True)"],
            schelp="""Flag determining whether to enable authenticated and encrypted jobs.""")

    scparam(cfg, ['option', 'cfg'],
            sctype='[file]',
            scope='job',
            shorthelp="Configuration manifest",
            switch="-cfg <file>",
            example=["cli: -cfg mypdk.json",
                     "api: chip.set('option', 'cfg', 'mypdk.json')"],
            schelp="""
            List of filepaths to JSON formatted schema configuration
            manifests. The files are read in automatically when using the
            command line application. In Python programs, JSON manifests
            can be merged into the current working manifest using the
            read_manifest() method.""")

    scparam(cfg, ['option', 'loglevel'],
            sctype='enum',
            enum=["info", "warning", "error", "critical", "debug"],
            pernode='optional',
            scope='job',
            defvalue='info',
            shorthelp="Logging level",
            switch="-loglevel <str>",
            example=[
                "cli: -loglevel info",
                "api: server.set('option', 'loglevel', 'info')"],
            schelp="""
            Provides explicit control over the level of debug logging printed.""")

    scparam(cfg, ['option', 'checkinterval'],
            sctype='int',
            defvalue=30,
            shorthelp="Interval for client",
            switch="-checkinterval <int>",
            example=["cli: -checkinterval 10",
                     "api: chip.set('option', 'checkinterval', 10)"],
            schelp="""
            Interval between checks to announce to clients""")

    return cfg


class ServerSchema(Schema):
    def __init__(self, cfg=None, manifest=None, logger=None):
        super().__init__(cfg=cfg,
                         manifest=manifest,
                         logger=logger)

    def _init_schema_cfg(self):
        return schema_cfg()



import siliconcompiler
from siliconcompiler.flows import lintflow, asicflow, asictopflow, synflow

from lambdapdk import freepdk45
from lambdapdk.freepdk45.libs import nangate45


####################################################
# Target Setup
####################################################
def setup(chip, syn_np=1, floorplan_np=1, physyn_np=1, place_np=1, cts_np=1, route_np=1,
          timing_np=1):
    '''
    FreePDK45 demo target
    '''

    asic_flow_args = {
        "syn_np": syn_np,
        "floorplan_np": floorplan_np,
        "physyn_np": physyn_np,
        "place_np": place_np,
        "cts_np": cts_np,
        "route_np": route_np
    }

    # 1. Load PDK, flow, libs combo
    chip.use(freepdk45)
    chip.use(nangate45)
    chip.use(lintflow)
    chip.use(asicflow, **asic_flow_args)
    chip.use(synflow, syn_np=syn_np, timing_np=timing_np)
    chip.use(asictopflow)

    # 2. Set flow and pdk
    chip.set('option', 'flow', 'asicflow', clobber=False)
    chip.set('option', 'pdk', 'freepdk45', clobber=False)
    chip.set('option', 'stackup', '10M', clobber=False)

    # 3. Select libraries
    chip.set('asic', 'logiclib', 'nangate45', clobber=False)

    # 4. Set project specific design choices
    chip.set('asic', 'delaymodel', 'nldm', clobber=False)
    chip.set('constraint', 'density', 40, clobber=False)
    chip.set('constraint', 'coremargin', 1, clobber=False)

    # 5. Timing corners
    corner = 'typical'
    chip.set('constraint', 'timing', 'worst', 'libcorner', corner, clobber=False)
    chip.set('constraint', 'timing', 'worst', 'pexcorner', corner, clobber=False)
    chip.set('constraint', 'timing', 'worst', 'mode', 'func', clobber=False)
    chip.set('constraint', 'timing', 'worst', 'check', ['setup', 'hold'], clobber=False)


#########################
if __name__ == "__main__":
    target = siliconcompiler.Chip('<target>')
    setup(target)
    target.write_manifest('freepdk45_demo.json')



# Copyright 2024 Silicon Compiler Authors. All Rights Reserved.

# Standard Modules
import sys

import siliconcompiler
from siliconcompiler.apps._common import UNSET_DESIGN
from siliconcompiler import SiliconCompilerError


###########################
def main():
    progname = "summarize"
    description = """
    ------------------------------------------------------------
    Utility script to print job summary from a manifest
    ------------------------------------------------------------
    """
    # Create a base chip class.
    chip = siliconcompiler.Chip(UNSET_DESIGN)

    # Read command-line inputs and generate Chip objects to run the flow on.
    try:
        chip.create_cmdline(progname,
                            description=description,
                            switchlist=['-cfg',
                                        '-loglevel'])
    except SiliconCompilerError:
        return 1
    except Exception as e:
        chip.logger.error(e)
        return 1

    design = chip.get('design')
    if design == UNSET_DESIGN:
        chip.logger.error('Design not loaded')
        return 1

    # Print Job Summary
    chip.summary(generate_image=False, generate_html=False)

    return 0


#########################
if __name__ == "__main__":
    sys.exit(main())



import os
import base64
import webbrowser
import subprocess
from siliconcompiler import utils

from siliconcompiler.report.utils import _collect_data, _find_summary_image
from siliconcompiler.tools._common import get_tool_task


def _generate_html_report(chip, flow, flowgraph_nodes, results_html):
    '''
    Generates an HTML based on the run
    '''

    # only report tool based steps functions
    for (step, index) in list(flowgraph_nodes):
        tool, task = get_tool_task(chip, step, '0', flow=flow)
        if tool == 'builtin':
            index = flowgraph_nodes.index((step, index))
            del flowgraph_nodes[index]

    schema = chip.schema.copy()
    schema.prune()
    pruned_cfg = schema.cfg
    if 'history' in pruned_cfg:
        del pruned_cfg['history']
    if 'library' in pruned_cfg:
        del pruned_cfg['library']

    layout_img = _find_summary_image(chip)

    img_data = None
    # Base64-encode layout for inclusion in HTML report
    if layout_img and os.path.isfile(layout_img):
        with open(layout_img, 'rb') as img_file:
            img_data = base64.b64encode(img_file.read()).decode('utf-8')

    nodes, errors, metrics, metrics_unit, metrics_to_show, reports = \
        _collect_data(chip, flow, flowgraph_nodes)

    # Hardcode the encoding, since there's a Unicode character in a
    # Bootstrap CSS file inlined in this template. Without this setting,
    # this write may raise an encoding error on machines where the
    # default encoding is not UTF-8.
    with open(results_html, 'w', encoding='utf-8') as wf:
        wf.write(utils.get_file_template('report/sc_report.j2').render(
            design=chip.design,
            nodes=nodes,
            errors=errors,
            metrics=metrics,
            metrics_unit=metrics_unit,
            reports=reports,
            manifest=chip.schema.cfg,
            pruned_cfg=pruned_cfg,
            metric_keys=metrics_to_show,
            img_data=img_data,
        ))

    chip.logger.info(f'Generated HTML report at {results_html}')


def _open_html_report(chip, results_html):
    try:
        webbrowser.get(results_html)
    except webbrowser.Error:
        # Python 'webbrowser' module includes a limited number of popular defaults.
        # Depending on the platform, the user may have defined their own with
        # $BROWSER.
        env_browser = os.getenv('BROWSER')
        if env_browser:
            subprocess.Popen([env_browser, os.path.relpath(results_html)])
        else:
            chip.logger.warning(f'Unable to open results page in web browser: {results_html}')



'''
sv2v converts SystemVerilog (IEEE 1800-2017) to Verilog
(IEEE 1364-2005), with an emphasis on supporting synthesizable
language constructs. The primary goal of this project is to
create a completely free and open-source tool for converting
SystemVerilog to Verilog. While methods for performing this
conversion already exist, they generally either rely on
commercial tools, or are limited in scope.

Documentation: https://github.com/zachjs/sv2v

Sources: https://github.com/zachjs/sv2v

Installation: https://github.com/zachjs/sv2v
'''

from siliconcompiler.tools.sv2v import convert


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    convert.setup(chip)
    return chip


def parse_version(stdout):
    # 0.0.7-130-g1aa30ea
    return '-'.join(stdout.split('-')[:-1])


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("sv2v.json")



import streamlit_antd_components as sac

from siliconcompiler.report.dashboard import state


def check_rerun():
    # Determine if node was modified
    if state.set_key(state.SELECTED_NODE, state.get_selected_node()):
        state.set_key(state.APP_RERUN, "Node")

    if state.get_key(state.APP_RERUN) == "Node":
        state.set_key(state.SELECT_TAB, "Node Information")
        state.del_key(state.TAB_STATE)
    elif state.get_key(state.APP_RERUN) == "File":
        state.set_key(state.SELECT_TAB, "File Viewer")
        state.del_key(state.TAB_STATE)
    else:
        state.set_key(state.SELECT_TAB, None)


def sac_tabs(tab_headings):
    index = 0

    if state.get_key(state.SELECT_TAB):
        for n, tab in enumerate(tab_headings):
            if state.get_key(state.SELECT_TAB) == tab.label:
                index = n
    elif state.get_key(state.TAB_INDEX) is not None:
        index = state.get_key(state.TAB_INDEX)

    tab_selected = sac.tabs(
        tab_headings,
        align='center',
        variant='outline',
        use_container_width=True,
        index=index,
        key=state.TAB_STATE)

    for n, tab in enumerate(tab_headings):
        if tab_selected == tab.label:
            state.set_key(state.TAB_INDEX, n)

    return tab_selected



from .summary_image import _generate_summary_image, _open_summary_image
from .html_report import _generate_html_report, _open_html_report
from .summary_table import _show_summary_table
from .dashboard import Dashboard

__all__ = [
    "_generate_summary_image",
    "_open_summary_image",
    "_generate_html_report",
    "_open_html_report",
    "_show_summary_table",
    "Dashboard"
]



import os


# TODO: this is a hack to get around design name requirement: since legal
# design names probably can't contain spaces, we can detect if it is unset.
UNSET_DESIGN = '  unset  '


def manifest_switches():
    '''
    Returns a list of manifest switches that can be used
    to find the manifest based on their values
    '''
    return ['-design',
            '-cfg',
            '-arg_step',
            '-arg_index',
            '-jobname']


def _get_manifests(cwd):
    manifests = {}

    def get_dirs(cwd):
        dirs = []
        for dirname in os.listdir(cwd):
            fullpath = os.path.join(cwd, dirname)
            if os.path.isdir(fullpath):
                dirs.append((dirname, fullpath))
        return dirs

    for _, buildpath in get_dirs(cwd):
        for design, designdir in get_dirs(buildpath):
            for jobname, jobdir in get_dirs(designdir):
                manifest = os.path.join(jobdir, f'{design}.pkg.json')
                if os.path.isfile(manifest):
                    manifests[(design, jobname, None, None)] = manifest
                for step, stepdir in get_dirs(jobdir):
                    for index, indexdir in get_dirs(stepdir):
                        manifest = os.path.join(indexdir, 'outputs', f'{design}.pkg.json')
                        if os.path.isfile(manifest):
                            manifests[(design, jobname, step, index)] = manifest
                        else:
                            manifest = os.path.join(indexdir, 'inputs', f'{design}.pkg.json')
                            if os.path.isfile(manifest):
                                manifests[(design, jobname, step, index)] = manifest

    organized_manifest = {}
    for (design, job, step, index), manifest in manifests.items():
        jobs = organized_manifest.setdefault(design, {})
        jobs.setdefault(job, {})[step, index] = os.path.abspath(manifest)

    return organized_manifest


def pick_manifest_from_file(chip, src_file, all_manifests):
    if src_file is None:
        return None

    if not os.path.exists(src_file):
        chip.logger.error(f'{src_file} cannot be found.')
        return None

    src_dir = os.path.abspath(os.path.dirname(src_file))
    for _, jobs in all_manifests.items():
        for _, nodes in jobs.items():
            for manifest in nodes.values():
                if src_dir == os.path.dirname(manifest):
                    return manifest

    return None


def pick_manifest(chip, src_file=None):
    all_manifests = _get_manifests(os.getcwd())

    manifest = pick_manifest_from_file(chip, src_file, all_manifests)
    if manifest:
        return manifest

    if chip.design == UNSET_DESIGN:
        if len(all_manifests) == 1:
            chip.set('design', list(all_manifests.keys())[0])
        else:
            chip.logger.error('Design name is not set')
            return None

    if chip.design not in all_manifests:
        chip.logger.error(f'Could not find manifest for {chip.design}')
        return None

    if chip.get('option', 'jobname') not in all_manifests[chip.design] and \
            len(all_manifests[chip.design]) != 1:
        chip.logger.error(f'Could not determine jobname for {chip.design}')
        return None

    jobname = chip.get('option', 'jobname')
    if chip.get('option', 'jobname') not in all_manifests[chip.design]:
        jobname = list(all_manifests[chip.design].keys())[0]

    if (None, None) in all_manifests[chip.design][jobname]:
        manifest = all_manifests[chip.design][jobname][None, None]
    else:
        # pick newest manifest
        manifest = list(sorted(all_manifests[chip.design][jobname].values(),
                               key=lambda file: os.stat(file).st_ctime))[-1]

    return manifest



from siliconcompiler.tools.builtin import _common
from siliconcompiler import flowgraph
from siliconcompiler.tools.builtin.builtin import set_io_files


def setup(chip):
    '''
    Selects the task with the minimum metric score from a list of inputs.

    Sequence of operation:

    1. Check list of input tasks to see if all metrics meets goals
    2. Check list of input tasks to find global min/max for each metric
    3. Select MIN value if all metrics are met.
    4. Normalize the min value as sel = (val - MIN) / (MAX - MIN)
    5. Return normalized value and task name

    Meeting metric goals takes precedence over compute metric scores.
    Only goals with values set and metrics with weights set are considered
    in the calculation.
    '''

    set_io_files(chip)


def _select_inputs(chip, step, index):
    inputs = _common._select_inputs(chip, step, index)

    score, sel_inputs = _common._minmax(chip, *inputs, op='minimum')

    if sel_inputs:
        chip.logger.info(f"Selected '{sel_inputs[0]}{sel_inputs[1]}' with score {score:.3f}")

    return sel_inputs


def _gather_outputs(chip, step, index):
    '''Return set of filenames that are guaranteed to be in outputs
    directory after a successful run of step/index.'''

    flow = chip.get('option', 'flow')

    in_nodes = chip.get('flowgraph', flow, step, index, 'input')
    in_task_outputs = [flowgraph._gather_outputs(chip, *node) for node in in_nodes]

    if len(in_task_outputs) > 0:
        return in_task_outputs[0].intersection(*in_task_outputs[1:])

    return []


def run(chip):
    return _common.run(chip)


def post_process(chip):
    _common.post_process(chip)



from siliconcompiler.tools import vivado
from siliconcompiler.tools.vivado import tool
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''Performs placement.'''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    vivado.setup_task(chip, task)

    design = chip.top()
    chip.set('tool', tool, 'task', task, 'input', f'{design}_checkpoint.dcp',
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', f'{design}_checkpoint.dcp',
             step=step, index=index)


def post_process(chip):
    vivado.post_process(chip)



import siliconcompiler
from siliconcompiler import SiliconCompilerError
from siliconcompiler.targets import freepdk45_demo


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    chip.use(freepdk45_demo)
    return setup(filetype='gds', showtools=chip._showtools, np=3)


###########################################################################
# Flowgraph Setup
############################################################################
def setup(flowname='showflow', filetype=None, screenshot=False, showtools=None, np=1):
    '''
    A flow to show the output files generated from other flows.

    Required settings for this flow are below:

    * filetype : Type of file to show

    Optional settings for this flow are below:

    * np : Number of parallel show jobs to launch
    * screenshot : true/false, indicate if this should be configured as a screenshot
    * showtools: dictionary of file extensions with the associated show and screenshot tasks
    '''

    flow = siliconcompiler.Flow(flowname)

    # Get required parameters first
    if not filetype:
        raise ValueError('filetype is a required argument')

    if not showtools:
        raise ValueError('showtools is a required argument')

    if filetype not in showtools:
        raise SiliconCompilerError(f'Show tool for {filetype} is not defined.')

    show_tool = showtools[filetype]

    stepname = 'show'
    if screenshot:
        stepname = 'screenshot'

    if stepname not in show_tool:
        raise SiliconCompilerError(f'{stepname} for {filetype} is not defined.')

    for idx in range(np):
        flow.node(flowname, stepname, show_tool[stepname], index=idx)

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



import os
from siliconcompiler.tools.opensta import setup as tool_setup
from siliconcompiler.tools.opensta import runtime_options as tool_runtime_options
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Report information about the timing libraries.
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    tool_setup(chip)

    chip.set('tool', tool, 'task', task, 'script', 'sc_report_libraries.tcl',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index)


################################
# Runtime options
################################
def runtime_options(chip):
    return tool_runtime_options(chip)



import altair
import math
import pandas
import streamlit

from siliconcompiler.report import report

from siliconcompiler.report.dashboard import state


def _get_report_chips():
    chips = []
    for job in state.get_chips():
        chips.append({'chip_object': state.get_chip(job), 'chip_name': job})
    return chips


def job_selector():
    """
    Displays a dataframe that can be edited to select specific jobs to include
    in the analysis.
    """
    jobs = state.get_chips()

    all_jobs = pandas.DataFrame({
        'job names': jobs,
        'selected jobs': [True] * len(jobs)
    })

    configuration = {
        'selected jobs': streamlit.column_config.CheckboxColumn(
            'Select runs',
            default=True)
    }

    with streamlit.popover('Select Jobs', use_container_width=True):
        selected_jobs = streamlit.data_editor(
            all_jobs,
            disabled=['job names'],
            use_container_width=True,
            hide_index=True,
            column_config=configuration)

        jobs = []
        for is_selected, job_name in zip(selected_jobs['selected jobs'].tolist(),
                                         selected_jobs['job names'].tolist()):
            if is_selected:
                jobs.append(job_name)

        state.set_key(state.GRAPH_JOBS, jobs)


def graph_count_selector():
    return streamlit.slider(
        'pick the number of graphs you want',
        1,
        10,
        1,
        label_visibility='collapsed')


def settings(metrics, nodes, graph_number):
    """
    Displays selectbox for metrics and nodes which informs the graph on what
    to display.

    Args:
        metrics (list) : A list of metrics that are set for all chips given in chips.
        nodes (list) : A list of nodes given in the form f'{step}{index}'
        graph_number (int) : The number of graphs there are. Used to create
            keys to distinguish selectboxes from each other.
    """
    metric_selector_col, node_selector_col, settings_col = \
        streamlit.columns(3, gap='small')

    with metric_selector_col:
        with streamlit.popover('Select a Metric', use_container_width=True):
            selected_metric = streamlit.selectbox(
                'Select a Metric',
                metrics,
                label_visibility='collapsed',
                key=f'graph-{graph_number}-metric-selection')

    with node_selector_col:
        with streamlit.popover('Select Nodes', use_container_width=True):
            selected_nodes = streamlit.multiselect(
                'Select a Node',
                nodes,
                label_visibility='collapsed',
                key=f'graph-{graph_number}-node-selection',
                default=nodes)

    with settings_col:
        with streamlit.popover('Settings', use_container_width=True):
            log_scale = streamlit.checkbox(
                "Log scale",
                False,
                help="Make the y-axis log scale",
                key=f'graph-{graph_number}-log-scale')

            transpose = streamlit.checkbox(
                "Transpose",
                False,
                help="Use nodes instead of jobs as the x-axis",
                key=f'graph-{graph_number}-transpose')

            chart_type = streamlit.selectbox(
                'Chart type',
                ['line', 'bar', 'point', 'tick'],
                label_visibility='collapsed',
                key=f'graph-{graph_number}-chart-selection')

    return selected_metric, selected_nodes, log_scale, transpose, chart_type


def graph(metrics, nodes, node_to_step_index_map, graph_number):
    metric, selected_nodes, log_scale, transpose, chart_type = \
        settings(metrics, nodes, graph_number)

    nodes_as_step_and_index = []
    for selected_node in selected_nodes:
        step, index = node_to_step_index_map[selected_node]
        nodes_as_step_and_index.append((step, index))

    if transpose:
        x_axis_label = 'nodes'
        color_label = 'runs'
    else:
        x_axis_label = 'runs'
        color_label = 'nodes'

    y_axis_label = metric

    data, metric_unit = report.get_chart_data(_get_report_chips(), metric, nodes_as_step_and_index)
    if metric_unit:
        y_axis_label = f'{metric}({metric_unit})'

    # Prepare plot data
    filtered_data = {
        x_axis_label: [],
        y_axis_label: [],
        color_label: []
    }

    if not nodes.empty:
        # filtering through data
        for job_name in state.get_key(state.GRAPH_JOBS):
            for step, index in data:
                filtered_data['runs'].append(job_name)
                filtered_data['nodes'].append(step + index)
                if job_name not in data[(step, index)].keys():
                    filtered_data[y_axis_label].append(None)
                else:
                    filtered_data[y_axis_label].append(data[(step, index)][job_name])

    # Setup chart
    x_axis = altair.X(x_axis_label, axis=altair.Axis(labelAngle=-75))

    y_axis = y_axis_label
    if log_scale and chart_type != 'bar':
        y_axis = altair.Y(y_axis_label, scale=altair.Scale(type="log"))

    color = color_label

    alt_chart = altair.Chart(pandas.DataFrame(filtered_data).dropna(), height=500)

    if chart_type == 'line':
        chart_mark = alt_chart.mark_line(point=True)
    elif chart_type == 'bar':
        chart_mark = alt_chart.mark_bar(point=True)
    elif chart_type == 'point':
        chart_mark = alt_chart.mark_circle(point=True)
    elif chart_type == 'tick':
        chart_mark = alt_chart.mark_tick(point=True)
    else:
        raise ValueError(f'{chart_type} not supported')

    chart = chart_mark.encode(
        x=x_axis,
        y=y_axis,
        color=color)

    streamlit.altair_chart(chart, use_container_width=True, theme='streamlit')


def viewer(metric_dataframe, node_to_step_index_map, metric_to_metric_unit_map):
    metrics = metric_dataframe.index.map(lambda x: metric_to_metric_unit_map[x])
    nodes = metric_dataframe.columns

    job_selector_col, graph_adder_col = streamlit.columns(2, gap='large')
    with job_selector_col:
        job_selector()
    with graph_adder_col:
        graphs = graph_count_selector()

    streamlit.divider()

    columns = 1 if graphs <= 1 else 2
    last_row_num = int(math.floor((graphs - 1) / columns)) * columns

    graph_number = 0
    graph_cols = streamlit.columns(columns, gap='large')
    while graph_number < graphs:
        with graph_cols[graph_number % columns]:
            graph(metrics, nodes, node_to_step_index_map, graph_number)
            if graph_number < last_row_num:
                streamlit.divider()
        graph_number += 1



class NodeStatus():
    # node waiting to run
    PENDING = 'pending'

    # node ready to run and waiting
    QUEUED = 'queued'

    # node running
    RUNNING = 'running'

    # node exit status
    SUCCESS = 'success'
    ERROR = 'error'
    SKIPPED = 'skipped'
    TIMEOUT = 'timeout'

    def is_done(status):
        return status in (
            NodeStatus.SUCCESS,
            NodeStatus.ERROR,
            NodeStatus.SKIPPED,
            NodeStatus.TIMEOUT
        )

    def is_running(status):
        return status in (
            NodeStatus.QUEUED,
            NodeStatus.RUNNING
        )

    def is_waiting(status):
        return status in (
            NodeStatus.PENDING,
        )

    def is_success(status):
        return status in (
            NodeStatus.SUCCESS,
            NodeStatus.SKIPPED
        )

    def is_error(status):
        return status in (
            NodeStatus.ERROR,
            NodeStatus.TIMEOUT
        )


###############################################################################
# Package Customization classes
###############################################################################
class SiliconCompilerError(Exception):
    ''' Minimal Exception wrapper used to raise sc runtime errors.
    '''
    def __init__(self, message, chip=None):
        super(Exception, self).__init__(message)

        if chip:
            logger = getattr(chip, 'logger', None)
            if logger:
                logger.error(message)



import os
import requests
import tarfile
import zipfile
from git import Repo, GitCommandError
from urllib.parse import urlparse
import importlib
import shutil
import re
from siliconcompiler import SiliconCompilerError
from siliconcompiler.utils import default_cache_dir, _resolve_env_vars
import json
from importlib.metadata import distributions, distribution
import functools
import fasteners
import time
from pathlib import Path
from io import BytesIO

from github import Github
import github.Auth


def get_cache_path(chip):
    cache_path = chip.get('option', 'cachedir')
    if cache_path:
        cache_path = chip.find_files('option', 'cachedir', missing_ok=True)
        if not cache_path:
            cache_path = os.path.join(chip.cwd, chip.get('option', 'cachedir'))
    if not cache_path:
        cache_path = default_cache_dir()

    return cache_path


def _path(chip, package, download_handler):
    if package in chip._packages:
        return chip._packages[package]

    # Initially try retrieving data source from schema
    data = {}
    data['path'] = chip.get('package', 'source', package, 'path')
    data['ref'] = chip.get('package', 'source', package, 'ref')
    if not data['path']:
        raise SiliconCompilerError(
            f'Could not find package source for {package} in schema. '
            'You can use register_source() to add it.', chip=chip)

    data['path'] = _resolve_env_vars(chip, data['path'])

    url = urlparse(data['path'])

    # check network drive for package data source
    if data['path'].startswith('file://') or os.path.exists(data['path']):
        path = os.path.abspath(data['path'].replace('file://', ''))
        chip.logger.info(f'Found {package} data at {path}')
        chip._packages[package] = path
        return path
    elif data['path'].startswith('python://'):
        path = path_from_python(chip, url.netloc)
        chip.logger.info(f'Found {package} data at {path}')
        chip._packages[package] = path
        return path

    # location of the python package
    cache_path = get_cache_path(chip)
    if not os.path.exists(cache_path):
        os.makedirs(cache_path, exist_ok=True)
    project_id = f'{package}-{data.get("ref")}'
    if url.scheme not in ['git', 'git+https', 'https', 'git+ssh', 'ssh'] or not project_id:
        raise SiliconCompilerError(
            f'Could not find data path in package {package}: {data["path"]}',
            chip=chip)

    data_path = os.path.join(cache_path, project_id)

    if download_handler:
        download_handler(chip,
                         package,
                         data,
                         url,
                         data_path,
                         os.path.join(cache_path, f'{project_id}.lock'))

    if os.path.exists(data_path):
        if package not in chip._packages:
            chip.logger.info(f'Saved {package} data to {data_path}')
            chip._packages[package] = data_path
        return data_path

    raise SiliconCompilerError(f'Extracting {package} data to {data_path} failed',
                               chip=chip)


def path(chip, package):
    """
    Compute data source data path
    Additionally cache data source data if possible
    Parameters:
        package (str): Name of the data source
    Returns:
        path: Location of data source on the local system
    """

    return _path(chip, package, __download_data)


def __download_data(chip, package, data, url, data_path, data_path_lock):
    data_lock = fasteners.InterProcessLock(data_path_lock)

    _aquire_data_lock(data_path, data_lock)

    # check cached package data source
    if os.path.exists(data_path):
        chip.logger.info(f'Found cached {package} data at {data_path}')
        if url.scheme in ['git', 'git+https', 'ssh', 'git+ssh']:
            try:
                repo = Repo(data_path)
                if repo.untracked_files or repo.index.diff("HEAD"):
                    chip.logger.warning('The repo of the cached data is dirty.')
                _release_data_lock(data_lock)
                chip._packages[package] = data_path
                return
            except GitCommandError:
                chip.logger.warning('Deleting corrupted cache data.')
                shutil.rmtree(path)
        else:
            _release_data_lock(data_lock)
            chip._packages[package] = data_path
            return

    # download package data source
    if url.scheme in ['git', 'git+https', 'ssh', 'git+ssh']:
        clone_synchronized(chip, package, data, data_path)
    elif url.scheme == 'https':
        extract_from_url(chip, package, data, data_path)

    _release_data_lock(data_lock)


def __get_filebased_lock(data_lock):
    base, _ = os.path.splitext(os.fsdecode(data_lock.path))
    return Path(f'{base}.sc_lock')


def _aquire_data_lock(data_path, data_lock):
    # Wait a maximum of 10 minutes for other processes to finish
    max_seconds = 10 * 60
    try:
        if data_lock.acquire(timeout=max_seconds):
            return
    except RuntimeError:
        # Fall back to file based locking method
        lock_file = __get_filebased_lock(data_lock)
        while (lock_file.exists()):
            if max_seconds == 0:
                raise SiliconCompilerError(f'Failed to access {data_path}.'
                                           f'Lock {lock_file} still exists.')
            time.sleep(1)
            max_seconds -= 1

        lock_file.touch()

        return

    raise SiliconCompilerError(f'Failed to access {data_path}. '
                               f'{data_lock.path} is still locked, if this is a mistake, '
                               'please delete it.')


def _release_data_lock(data_lock):
    # Check if file based locking method was used
    lock_file = __get_filebased_lock(data_lock)
    if lock_file.exists():
        lock_file.unlink(missing_ok=True)
        return

    data_lock.release()


def clone_synchronized(chip, package, data, data_path):
    url = urlparse(data['path'])
    try:
        clone_from_git(chip, package, data, data_path)
    except GitCommandError as e:
        if 'Permission denied' in repr(e):
            if url.scheme in ['ssh', 'git+ssh']:
                chip.logger.error('Failed to authenticate. Please setup your git ssh.')
            elif url.scheme in ['git', 'git+https']:
                chip.logger.error('Failed to authenticate. Please use a token or ssh.')
        else:
            chip.logger.error(str(e))


def clone_from_git(chip, package, data, repo_path):
    url = urlparse(data['path'])
    if url.scheme in ['git', 'git+https'] and url.username:
        chip.logger.warning('Your token is in the data source path and will be stored in the '
                            'schema. If you do not want this set the env variable GIT_TOKEN '
                            'or use ssh for authentication.')
    if url.scheme in ['git+ssh', 'ssh']:
        chip.logger.info(f'Cloning {package} data from {url.netloc}:{url.path[1:]}')
        # Git requires the format git@github.com:org/repo instead of git@github.com/org/repo
        repo = Repo.clone_from(f'{url.netloc}:{url.path[1:]}',
                               repo_path,
                               recurse_submodules=True)
    else:
        if os.environ.get('GIT_TOKEN') and not url.username:
            url = url._replace(netloc=f'{os.environ.get("GIT_TOKEN")}@{url.hostname}')
        url = url._replace(scheme='https')
        chip.logger.info(f'Cloning {package} data from {url.geturl()}')
        repo = Repo.clone_from(url.geturl(), repo_path, recurse_submodules=True)
    chip.logger.info(f'Checking out {data["ref"]}')
    repo.git.checkout(data["ref"])
    for submodule in repo.submodules:
        submodule.update(init=True)


def extract_from_url(chip, package, data, data_path):
    url = urlparse(data['path'])
    data_url = data.get('path')
    headers = {}
    if os.environ.get('GIT_TOKEN') or url.username:
        headers['Authorization'] = f'token {os.environ.get("GIT_TOKEN") or url.username}'
    if "github" in data_url:
        headers['Accept'] = 'application/octet-stream'
    data_url = data['path']
    if data_url.endswith('/'):
        data_url = f"{data_url}{data['ref']}.tar.gz"
    chip.logger.info(f'Downloading {package} data from {data_url}')
    response = requests.get(data_url, stream=True, headers=headers)
    if not response.ok:
        raise SiliconCompilerError(f'Failed to download {package} data source.', chip=chip)

    fileobj = BytesIO(response.content)
    try:
        with tarfile.open(fileobj=fileobj, mode='r|gz') as tar_ref:
            tar_ref.extractall(path=data_path)
    except tarfile.ReadError:
        fileobj.seek(0)
        # Try as zip
        with zipfile.ZipFile(fileobj) as zip_ref:
            zip_ref.extractall(path=data_path)

    if 'github' in url.netloc and len(os.listdir(data_path)) == 1:
        # Github inserts one folder at the highest level of the tar file
        # this compensates for this behavior
        gh_url = urlparse(data_url)

        repo = gh_url.path.split('/')[2]

        ref = gh_url.path.split('/')[-1]
        if repo.endswith('.git'):
            ref = data['ref']
        elif ref.endswith('.tar.gz'):
            ref = ref[0:-7]
        elif ref.endswith('.tgz'):
            ref = ref[0:-4]
        else:
            ref = ref.split('.')[0]

        if ref.startswith('v'):
            ref = ref[1:]

        github_folder = f"{repo}-{ref}"

        if github_folder in os.listdir(data_path):
            # This moves all files one level up
            git_path = os.path.join(data_path, github_folder)
            for data_file in os.listdir(git_path):
                shutil.move(os.path.join(git_path, data_file), data_path)
            os.removedirs(git_path)


def path_from_python(chip, python_package, append_path=None):
    try:
        module = importlib.import_module(python_package)
    except:  # noqa E722
        raise SiliconCompilerError(f'Failed to import {python_package}.', chip=chip)

    python_path = os.path.dirname(module.__file__)
    if append_path:
        if isinstance(append_path, str):
            append_path = [append_path]
        python_path = os.path.join(python_path, *append_path)

    python_path = os.path.abspath(python_path)

    return python_path


def is_python_module_editable(module_name):
    dist_map = __get_python_module_mapping()
    dist = dist_map[module_name][0]

    is_editable = False
    for f in distribution(dist).files:
        if f.name == 'direct_url.json':
            info = None
            with open(f.locate(), 'r') as f:
                info = json.load(f)

            if "dir_info" in info:
                is_editable = info["dir_info"].get("editable", False)

    return is_editable


def register_python_data_source(chip,
                                package_name,
                                python_module,
                                alternative_path,
                                alternative_ref=None,
                                python_module_path_append=None):
    '''
    Helper function to register a python module as data source with an alternative in case
    the module is not installed in an editable state
    '''
    # check if installed in an editable state
    if is_python_module_editable(python_module):
        if python_module_path_append:
            path = path_from_python(chip, python_module, append_path=python_module_path_append)
        else:
            path = f"python://{python_module}"
        ref = None
    else:
        path = alternative_path
        ref = alternative_ref

    chip.register_source(name=package_name,
                         path=path,
                         ref=ref)


def register_private_github_data_source(chip,
                                        package_name,
                                        repository,
                                        release,
                                        artifact):
    gh = Github(auth=github.Auth.Token(__get_github_auth_token(package_name)))
    repo = gh.get_repo(repository)

    if not release:
        release = repo.get_latest_release().tag_name

    url = None
    for repo_release in repo.get_releases():
        if repo_release.tag_name == release:
            for asset in repo_release.assets:
                if asset.name == artifact:
                    url = asset.url

    if not url:
        raise ValueError(f'Unable to find release asset: {repository}/{release}/{artifact}')

    chip.register_source(
        package_name,
        path=url,
        ref=release)


def __get_github_auth_token(package_name):
    token_name = package_name.upper()
    for tok in ('#', '$', '&', '-', '=', '!', '/'):
        token_name = token_name.replace(tok, '')

    search_env = (
        f'GITHUB_{token_name}_TOKEN',
        'GITHUB_TOKEN',
        'GIT_TOKEN'
    )

    token = None
    for env in search_env:
        token = os.environ.get(env, None)

        if token:
            break

    if not token:
        raise ValueError('Unable to determine authorization token for GitHub, '
                         f'please set one of the following environmental variables: {search_env}')

    return token


@functools.lru_cache(maxsize=1)
def __get_python_module_mapping():
    mapping = {}

    for dist in distributions():
        dist_name = None
        if hasattr(dist, 'name'):
            dist_name = dist.name
        else:
            metadata = dist.read_text('METADATA')
            if metadata:
                find_name = re.compile(r'Name: (.*)')
                for data in metadata.splitlines():
                    group = find_name.findall(data)
                    if group:
                        dist_name = group[0]
                        break

        if not dist_name:
            continue

        provides = dist.read_text('top_level.txt')
        if provides:
            for module in dist.read_text('top_level.txt').split():
                mapping.setdefault(module, []).append(dist_name)

    return mapping



from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Generate a bitstream for the ICE40 FPGA
    '''
    tool = 'icepack'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    clobber = False
    design = chip.top()

    chip.set('tool', tool, 'exe', tool)

    chip.set('tool', tool, 'task', task, 'option', "", step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'input', f'{design}.asc', step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', f'{design}.bit', step=step, index=index)



# Copyright 2020 Silicon Compiler Authors. All Rights Reserved.

# Standard Modules
import os
import sys

import siliconcompiler
from siliconcompiler.utils import get_default_iomap
from siliconcompiler.targets import skywater130_demo
from siliconcompiler import SiliconCompilerError


def _infer_designname(chip):
    topfile = None
    sourcesets = chip.getkeys('input')
    for sourceset in reversed(('rtl', 'hll')):
        if sourceset in sourcesets:
            sourcesets.remove(sourceset)
            sourcesets.insert(0, sourceset)
    for sourceset in sourcesets:
        for filetype in chip.getkeys('input', sourceset):
            all_vals = chip.schema._getvals('input', sourceset, filetype)
            if all_vals:
                # just look at first value
                sources, _, _ = all_vals[0]
                # grab first source
                topfile = sources[0]
                break
        if topfile:
            break

    if not topfile:
        return None

    root = os.path.basename(topfile)
    while True:
        root, ext = os.path.splitext(root)
        if not ext:
            break

    return root


###########################
def main():
    progname = "sc"
    description = """
    ------------------------------------------------------------
    SiliconCompiler is an open source compiler framework that
    aims to enable automated translation from source code to
    silicon.

    The sc program includes the following steps.

    1. Read command line arguments
    2. If not set, 'design' is set to base of first source file.
    3. If not set, 'target' is set to 'skywater130_demo'.
    4. Run compilation
    5. Display summary

    Sources: https://github.com/siliconcompiler/siliconcompiler
    ------------------------------------------------------------
    """
    # TODO: this is a hack to get around design name requirement: since legal
    # design names probably can't contain spaces, we can detect if it is unset.
    UNSET_DESIGN = '  unset  '

    # Create a base chip class.
    chip = siliconcompiler.Chip(UNSET_DESIGN)

    # Read command-line inputs and generate Chip objects to run the flow on.
    try:
        args = chip.create_cmdline(progname,
                                   description=description,
                                   input_map=get_default_iomap())
    except SiliconCompilerError:
        return 1
    except Exception as e:
        chip.logger.error(e)
        return 1

    # Set design if none specified
    if chip.get('design') == UNSET_DESIGN:
        topmodule = _infer_designname(chip)

        if not topmodule:
            chip.logger.error('Invalid arguments: either specify -design or provide sources.')
            return 1

        chip.set('design', topmodule)

    # Set demo target if none specified
    if 'target' not in args or not args['target']:
        chip.use(skywater130_demo)

    try:
        # Run flow
        chip.run()

        # Print Job Summary
        chip.summary()
    except SiliconCompilerError:
        return 1
    except Exception as e:
        chip.logger.error(e)
        return 1

    return 0


#########################
if __name__ == "__main__":
    sys.exit(main())



from siliconcompiler import Chip, Flow

from siliconcompiler.tools.klayout import operations
from siliconcompiler.tools.klayout import screenshot
from siliconcompiler.tools.montage import tile


def make_docs(chip):
    chip.set('input', 'layout', 'gds', 'test')
    chip.set('tool', 'klayout', 'task', 'screenshot', 'var', 'xbins', 2)
    chip.set('tool', 'klayout', 'task', 'screenshot', 'var', 'ybins', 2)
    chip.set('tool', 'montage', 'task', 'tile', 'var', 'xbins', 2)
    chip.set('tool', 'montage', 'task', 'tile', 'var', 'ybins', 2)
    return setup()


def setup(flowname='screenshotflow'):
    '''
    Flow to generate a high resolution design image from a GDS or OAS file.

    The 'screenshotflow' includes the stages below.

    * **prepare**: Prepare the stream file, such as flattening design, removing layers, and merging shapes
    * **screenshot**: Generate a set of screenshots tiled across the design
    * **merge**: Merge tiled images into a single image
    '''  # noqa E501

    pipe = [
        ('prepare', operations),
        ('screenshot', screenshot),
        ('merge', tile)
    ]

    flow = Flow(flowname)

    prevstep = None
    for step, task in pipe:
        flow.node(flowname, step, task)
        if prevstep:
            flow.edge(flowname, prevstep, step)
        prevstep = step

    return flow


##################################################
if __name__ == "__main__":
    chip = Chip('design')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



import fnmatch
import pandas
import os
from siliconcompiler import Schema
from siliconcompiler.report import utils
from siliconcompiler.flowgraph import nodes_to_execute
from siliconcompiler.tools._common import get_tool_task


def make_metric_dataframe(chip):
    '''
    Returns a pandas dataframe to display in the data metrics table. All nodes
    (steps and indices) are included on the x-axis while the metrics tracked
    are on the y-axis. The y-axis row headers are in the form of a tuple where
    the first element is the metric tracked and the second element is the unit.

    Args:
        chip (Chip) : The chip object that contains the schema read from.

    Example:
        >>> make_metric_dataframe(chip)
        Returns pandas dataframe of tracked metrics.
    '''
    nodes, errors, metrics, metrics_unit, metrics_to_show, reports = utils._collect_data(chip)
    # converts from 2d dictionary to pandas DataFrame, transposes so
    # orientation is correct, and filters based on the metrics we track
    data = (pandas.DataFrame.from_dict(metrics, orient='index').transpose())
    data = data.loc[metrics_to_show]
    # include metrics_unit
    data.index = data.index.map(lambda x: (x, metrics_unit[x]))
    return data


def get_flowgraph_nodes(chip, step, index):
    '''
    Returns a dictionary to display in the data metrics table. One node
    (step and index) is included on the x-axis while all the metrics tracked
    are on the y-axis. Removes all key value pairs where the value is None.

    Args:
        chip (Chip) : The chip object that contains the schema read from.
        step (string) : Step of node.
        index (string) : Index of node.

    Example:
        >>> get_flowgraph_nodes(chip, [(import, 0), (syn, 0)])
        Returns pandas dataframe of tracked metrics.
    '''
    nodes = {}
    tool, task = get_tool_task(chip, step, index)
    if tool is not None:
        nodes['tool'] = tool
    if task is not None:
        nodes['task'] = task
    for key in chip.getkeys('record'):
        if chip.get('record', key, field='pernode') == 'never':
            value = chip.get('record', key)
        else:
            value = chip.get('record', key, step=step, index=index)
        if value is not None:
            if key == 'inputnode':
                value = ", ".join([f'{step}{index}' for step, index in value])
            nodes[key] = str(value)
    return nodes


def get_flowgraph_edges(chip):
    '''
    Returns a dicitionary where each key is one node, a tuple in the form
    (step, index) and the value of each key is a set of tuples in the form
    (step, index). The value of each key represents all the nodes that are
    inputs to the key node.

    Args:
        chip (Chip) : The chip object that contains the schema read from.

    Example:
        >>> get_flowgraph_edges(chip)
        Returns dictionary where the values of the keys are the edges.
    '''
    flowgraph_edges = {}
    flow = chip.get('option', 'flow')
    for step in chip.getkeys('flowgraph', flow):
        for index in chip.getkeys('flowgraph', flow, step):
            flowgraph_edges[step, index] = set()
            for in_step, in_index in chip.get('flowgraph', flow, step, index, 'input'):
                flowgraph_edges[step, index].add((in_step, in_index))
    return flowgraph_edges


def make_manifest_helper(manifest_subsect, modified_manifest_subsect):
    '''
    Function is a helper function to make_manifest. It mutates the input json.

    Args:
        manifest_subsect (dict) : Represents a subset of the original manifest.
        modified_manifest_subsect (dict) : Represents a subset of the original
            manifest, modified for readability.

    Example:
        >>> make_manifest_helper(manifest_subsection, {})
        Mutates second paramaeter to remove simplify leaf nodes and remove
        'default' nodes.
    '''

    def build_leaf(manifest_subsect):
        if manifest_subsect['pernode'] == 'never':
            if Schema.GLOBAL_KEY in manifest_subsect['node'] and \
                    Schema.GLOBAL_KEY in manifest_subsect['node'][Schema.GLOBAL_KEY]:
                value = manifest_subsect['node'][Schema.GLOBAL_KEY][Schema.GLOBAL_KEY]['value']
            else:
                value = manifest_subsect['node']['default']['default']['value']
            return value
        else:
            nodes = manifest_subsect['node']
            node_values = {}
            for step in nodes:
                if step == 'default' or step == Schema.GLOBAL_KEY:
                    value = nodes[step][step]['value']
                    node_values[step] = value
                else:
                    for index in nodes[step]:
                        value = nodes[step][index]['value']
                        if value is None:
                            continue
                        if index == 'default' or index == Schema.GLOBAL_KEY:
                            node_values[step] = value
                        else:
                            node_values[step + index] = value
            return node_values

    if Schema._is_leaf(manifest_subsect):
        if manifest_subsect['pernode'] == 'never':
            if Schema.GLOBAL_KEY in manifest_subsect['node']:
                value = manifest_subsect['node'][Schema.GLOBAL_KEY][Schema.GLOBAL_KEY]['value']
            else:
                value = manifest_subsect['node']['default']['default']['value']
            modified_manifest_subsect['value'] = value
        else:
            nodes = manifest_subsect['node']
            for step in nodes:
                if step == 'default' or step == Schema.GLOBAL_KEY:
                    value = nodes[step][step]['value']
                    modified_manifest_subsect[step] = value
                else:
                    for index in nodes[step]:
                        value = nodes[step][index]['value']
                        if value is None:
                            continue
                        if index == 'default' or index == Schema.GLOBAL_KEY:
                            modified_manifest_subsect[step] = value
                        else:
                            modified_manifest_subsect[step + index] = value

    for key, key_dict in manifest_subsect.items():
        if key != 'default':
            if Schema._is_leaf(key_dict):
                modified_manifest_subsect[key] = build_leaf(key_dict)
            else:
                modified_manifest_subsect[key] = {}
                make_manifest_helper(key_dict, modified_manifest_subsect[key])


def make_manifest(chip):
    '''
    Returns a dictionary of dictionaries/json

    Args:
        chip (Chip) : The chip object that contains the schema read from.

    Example:
        >>> make_manifest(chip)
        Returns tree/json of manifest.
    '''
    manifest = chip.schema.cfg
    modified_manifest = {}
    make_manifest_helper(manifest, modified_manifest)
    return modified_manifest


def get_flowgraph_path(chip):
    '''
    Returns a set of all the nodes in the 'winning' path.

    Args:
        chip (Chip) : The chip object that contains the schema read from.

    Example:
        >>> get_flowgraph_path(chip)
        Returns the "winning" path for that job.
    '''
    flow = chip.get('option', 'flow')
    return utils._get_flowgraph_path(chip, flow, nodes_to_execute(chip))


def search_manifest_keys(manifest, key):
    '''
    Function is a recursive helper to search_manifest, more info there.

    Args:
        manifest (dictionary) : A dictionary representing the manifest.
        key (string) : Searches all keys for partial matches on this string.
    '''
    filtered_manifest = {}
    for dict_key in manifest:
        if fnmatch.fnmatch(dict_key, key):
            filtered_manifest[dict_key] = manifest[dict_key]
        elif isinstance(manifest[dict_key], dict):
            result = search_manifest_keys(manifest[dict_key], key)
            if result:  # result is non-empty
                filtered_manifest[dict_key] = result
    return filtered_manifest


def search_manifest_values(manifest, value):
    '''
    Function is a recursive helper to search_manifest, more info there.

    Args:
        manifest (dictionary) : A dicitionary representing the manifest.
        value (string) : Searches all values for partial matches on this
            string.
    '''
    filtered_manifest = {}
    for key in manifest:
        if isinstance(manifest[key], dict):
            result = search_manifest_values(manifest[key], value)
            if result:  # result is non-empty
                filtered_manifest[key] = result
        else:
            if manifest[key] is None:
                continue

            if isinstance(manifest[key], (list, tuple)):
                if fnmatch.filter([str(v) for v in manifest[key]], value):
                    filtered_manifest[key] = manifest[key]
            else:
                if fnmatch.fnmatch(str(manifest[key]), value):
                    filtered_manifest[key] = manifest[key]
    return filtered_manifest


def search_manifest(manifest, key_search=None, value_search=None):
    '''
    Returns the same structure as make_manifest, but it is filtered by partial
    matches by keys or values. If both key_search and value_search are None,
    the original manifest is returned.

    Args:
        manifest (dictionary) : A dicitionary representing the manifest.
        key_search (string) : Searches all keys for partial matches on this
            string
        value_search(string) : Searches all values for partial matches
            on this string.

    Example:
        >>> search_manifest(jsonDict, key_search='input', value_search='v')
        Returns a filtered version of jsonDict where each path contains at
        least one key that contains the substring input and has values that
        contain v.
    '''
    return_manifest = manifest
    if key_search:
        if '*' not in key_search or '?' not in key_search:
            key_search = f'*{key_search}*'
        return_manifest = search_manifest_keys(return_manifest, key_search)
    if value_search:
        if '*' not in value_search or '?' not in value_search:
            value_search = f'*{value_search}*'
        return_manifest = search_manifest_values(return_manifest, value_search)
    return return_manifest


def get_total_manifest_key_count(manifest):
    '''
    Returns (int) the number of keys

    Args:
        manifest (dictionary) : A dicitionary representing the manifest.
        acc (int) : An accumulator of the current number of folders and files.
    '''
    acc = len(manifest)
    for dictKeys in manifest:
        if isinstance(manifest[dictKeys], dict):
            acc += get_total_manifest_key_count(manifest[dictKeys])
    return acc


def get_metrics_source(chip, step, index):
    '''
    Returns a dictionary where the keys are files in the logs and reports for
    a given step and index. The values are a list of the metrics that come from
    that file. If a file is not in the dictionary, that implies that no metrics
    come from it.

    Args:
        chip (Chip) : The chip object that contains the schema read from.
        step (string) : Step of node.
        index (string) : Index of node.
    '''
    file_to_metric = {}
    tool, task = get_tool_task(chip, step, index)
    if not chip.valid('tool', tool, 'task', task, 'report'):
        return file_to_metric

    metrics = chip.getkeys('tool', tool, 'task', task, 'report')

    for metric in metrics:
        sources = chip.get('tool', tool, 'task', task, 'report', metric, step=step, index=index)
        for source in sources:
            if source in file_to_metric:
                file_to_metric[source].append(metric)
            else:
                file_to_metric[source] = [metric]
    return file_to_metric


def get_files(chip, step, index):
    '''
    Returns a list of 3-tuple that contain the path name of how to get to that
    folder, the subfolders of that directory, and it's files. The list is
    ordered by layer of directory.

    Args:
        chip (Chip) : The chip object that contains the schema read from.
        step (string) : Step of node.
        index (string) : Index of node.
    '''
    # could combine filters, but slightly more efficient to separate them
    # Is remaking the list with sets instead of list worth it?
    logs_and_reports = []
    all_paths = os.walk(chip.getworkdir(step=step, index=index))
    for path_name, folders, files in all_paths:
        logs_and_reports.append((path_name, set(folders), set(files)))
    return logs_and_reports


def get_chart_data(chips, metric, nodes):
    '''
    Returns returns a a tuple where the first element is a 2d dictionary of
    data points, following the forms {step+index: {chip_name: value}} where
    each dictionary can have many keys. The second element is a string that represents the unit.

    Args:
        chips (list) : A list of dictionaries with the form
            {'chip_object': chip, 'chip_name': name}.
        metric (string) : The metric that the user is searching.
        nodes (list) : A list of dictionaries with the form (step, index).
    '''
    metric_units = set()  # the set of all units for this metric (hopefully, it's length is 0 or 1)
    metric_datapoints = {}
    metric_unit = ''
    for chip_and_chip_name in chips:
        chip = chip_and_chip_name['chip_object']
        chip_name = chip_and_chip_name['chip_name']
        nodes_list, errors, metrics, metrics_unit, metrics_to_show, reports = \
            utils._collect_data(chip, format_as_string=False)
        if metric in metrics_unit:
            metric_unit = metrics_unit[metric]
            metric_units.add(metric_unit)
        for node in nodes:
            if node not in metrics:
                continue
            value = metrics[node][metric]
            if value is None:
                continue
            if node in metric_datapoints:
                metric_datapoints[node][chip_name] = value
            else:
                metric_datapoints[node] = {chip_name: value}
    if len(metric_units) > 1:
        raise ValueError('Not all measurements were made with the same units')
    return metric_datapoints, metric_unit



'''
Chisel is a hardware design language that facilitates advanced circuit
generation and design reuse for both ASIC and FPGA digital logic designs.
Chisel adds hardware construction primitives to the Scala programming
language, providing designers with the power of a modern programming
language to write complex, parameterizable circuit generators that produce
synthesizable Verilog.

Documentation: https://www.chisel-lang.org/docs

Sources: https://github.com/chipsalliance/chisel

Installation: The Chisel plugin relies on having the Scala Build Tool (sbt)
installed. Instructions: https://www.scala-sbt.org/download.html.
'''

from siliconcompiler.tools.chisel import convert


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    convert.setup(chip)
    return chip


def parse_version(stdout):
    # sbt version in this project: 1.5.5
    # sbt script version: 1.5.5

    for line in stdout.splitlines():
        line = line.strip()
        if 'sbt script version:' in line:
            return line.split()[-1]

    return None



import siliconcompiler


def setup():
    '''Subset of OH! library tapeout checklist.

    https://github.com/aolofsson/oh/blob/master/docs/tapeout_checklist.md
    '''
    standard = 'oh_tapeout'

    checklist = siliconcompiler.Checklist(standard)

    # Automated
    checklist.set('checklist', standard, 'drc_clean', 'description',
                  'Is block DRC clean?')
    checklist.set('checklist', standard, 'drc_clean', 'criteria', 'drcs==0')

    checklist.set('checklist', standard, 'lvs_clean', 'description',
                  'Is block LVS clean?')
    checklist.set('checklist', standard, 'lvs_clean', 'criteria', 'drcs==0')

    checklist.set('checklist', standard, 'setup_time', 'description',
                  'Setup time met?')
    checklist.set('checklist', standard, 'setup_time', 'criteria', 'setupslack>=0')

    checklist.set('checklist', standard, 'errors_warnings', 'description',
                  'Are all EDA warnings/errors acceptable?')
    checklist.set('checklist', standard, 'errors_warnings', 'criteria',
                  ['errors==0', 'warnings==0'])

    # Manual
    checklist.set('checklist', standard, 'spec', 'description',
                  'Is there a written specification?')

    return checklist


##################################################
if __name__ == "__main__":
    checklist = setup(siliconcompiler.Chip('<checklist>'))
    checklist.write_manifest(f"{checklist.top()}.json")



'''
Surelog is a SystemVerilog pre-processor, parser, elaborator,
and UHDM compiler that provides IEEE design and testbench
C/C++ VPI and a Python AST API.

Documentation: https://github.com/chipsalliance/Surelog

Sources: https://github.com/chipsalliance/Surelog

Installation: https://github.com/chipsalliance/Surelog
'''

import surelog
from siliconcompiler.tools._common import get_tool_task


################################
# Setup Tool (pre executable)
################################
def setup(chip):
    ''' Sets up default settings common to running Surelog.
    '''

    tool = 'surelog'
    # Nothing in this method should rely on the value of 'step' or 'index', but they are used
    # as schema keys in some important places, so we still need to fetch them.
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    exe = tool
    _, task = get_tool_task(chip, step, index)

    is_docker = chip.get('option', 'scheduler', 'name', step=step, index=index) == 'docker'
    if not is_docker:
        exe = surelog.get_bin()
    else:
        exe = surelog.get_bin('linux')

    # Standard Setup
    chip.set('tool', tool, 'exe', exe)
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=1.51', clobber=False)

    # We package SC wheels with a precompiled copy of Surelog installed to
    # tools/surelog/bin. If the user doesn't have Surelog installed on their
    # system path, set the path to the bundled copy in the schema.
    if not surelog.has_system_surelog() and not is_docker:
        chip.set('tool', tool, 'path', surelog.get_path(), clobber=False)

    # Log file parsing
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', r'^\[WRN:',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', r'^\[(ERR|FTL|SNT):',
             step=step, index=index, clobber=False)

    for warning in chip.get('tool', tool, 'task', task, 'warningoff', step=step, index=index):
        chip.add('tool', tool, 'regex', step, index, 'warnings', f'-v {warning}',
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'enable_lowmem',
             'true/false, when true instructs Surelog to minimize its maximum memory usage.',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'enable_lowmem', 'false',
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'enable_lowmem']),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'disable_write_cache',
             'true/false, when true instructs Surelog to not write to its cache.',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'disable_write_cache', 'false',
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'disable_write_cache']),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'disable_info',
             'true/false, when true instructs Surelog to not log infos.',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'disable_info', 'false',
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'disable_info']),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'disable_note',
             'true/false, when true instructs Surelog to not log notes.',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'disable_note', 'false',
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'disable_note']),
             step=step, index=index)


def parse_version(stdout):
    # Surelog --version output looks like:
    # VERSION: 1.13
    # BUILT  : Nov 10 2021

    # grab version # by splitting on whitespace
    return stdout.split()[1]


def runtime_options(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # Command-line options.
    options = []

    # With newer versions of Surelog (at least 1.35 and up), this option is
    # necessary to make bundled versions work.
    # TODO: why?
    options.append('-nocache')

    lowmem = chip.get('tool', tool, 'task', task, 'var', 'enable_lowmem', step=step, index=index)
    if lowmem == ['true']:
        options.append('-lowmem')

    no_write_cache = chip.get('tool', tool, 'task', task, 'var', 'disable_write_cache', step=step,
                              index=index)
    if no_write_cache == ['true']:
        options.append('-nowritecache')

    no_info = chip.get('tool', tool, 'task', task, 'var', 'disable_info', step=step, index=index)
    if no_info == ['true']:
        options.append('-noinfo')

    no_note = chip.get('tool', tool, 'task', task, 'var', 'disable_note', step=step, index=index)
    if no_note == ['true']:
        options.append('-nonote')

    return options



import pya
import importlib.util as importlib_util
import os
import shutil
import sys


def get_streams(schema):
    sc_step = schema.get('arg', 'step')
    sc_index = schema.get('arg', 'index')
    sc_flow = schema.get('option', 'flow')
    sc_task = schema.get('flowgraph', sc_flow, sc_step, sc_index, 'task')

    streams = ('gds', 'oas')
    if schema.valid('tool', 'klayout', 'task', sc_task, 'var', 'stream'):
        sc_stream = schema.get('tool', 'klayout', 'task', sc_task, 'var', 'stream',
                               step=sc_step, index=sc_index)[0]
    else:
        sc_stream = streams[0]
    return [sc_stream, *[s for s in streams if s != sc_stream]]


def technology(design, schema):
    from tools._common.asic import get_libraries

    sc_step = schema.get('arg', 'step')
    sc_index = schema.get('arg', 'index')
    sc_pdk = schema.get('option', 'pdk')

    if schema.valid('option', 'stackup'):
        sc_stackup = schema.get('option', 'stackup')
    else:
        sc_stackup = schema.get('pdk', sc_pdk, 'stackup')[0]

    logiclibs = schema.get('asic', 'logiclib', step=sc_step, index=sc_index)
    if not logiclibs:
        sc_libtype = schema.get('option', 'var', 'klayout_libtype')[0]
    else:
        sc_mainlib = logiclibs[0]
        sc_libtype = schema.get('library', sc_mainlib, 'asic', 'libarch',
                                step=sc_step, index=sc_index)

    sc_libs = []
    sc_libs += get_libraries(schema, 'logic')
    sc_libs += get_libraries(schema, 'macro')

    local_files = {
        'lyt': f'inputs/{design}.lyt',
        'lyp': f'inputs/{design}.lyp'
    }

    tech = pya.Technology.create_technology('sc_tech')
    # Load technology file
    tech_file = None
    if os.path.exists(local_files['lyt']):
        tech_file = local_files['lyt']
    elif schema.valid('pdk', sc_pdk, 'layermap', 'klayout', 'def', 'klayout', sc_stackup):
        tech_file = schema.get('pdk', sc_pdk, 'layermap', 'klayout', 'def', 'klayout', sc_stackup)
        if tech_file:
            tech_file = tech_file[0]

    if tech_file:
        print(f"[INFO] Loading technology file: {tech_file}")
        tech.load(tech_file)
    else:
        tech.name = f'{sc_pdk} for {sc_stackup}'
        tech.description = f'{", ".join(sc_libs)}'

    if schema.valid('pdk', sc_pdk, 'var', 'klayout', 'units', sc_stackup):
        units = float(schema.get('pdk', sc_pdk, 'var', 'klayout', 'units', sc_stackup)[0])
        tech.dbu = units
    print(f"[INFO] Technology database units are: {tech.dbu}um")

    lefs = []

    foundry_lef = schema.get('pdk', sc_pdk, 'aprtech', 'klayout', sc_stackup, sc_libtype, 'lef')
    if foundry_lef:
        foundry_lef = foundry_lef[0]
        lefs.append(foundry_lef)
    else:
        foundry_lef = None

    for lib in sc_libs:
        lefs.extend(schema.get('library', lib, 'output', sc_stackup, 'lef',
                               step=sc_step, index=sc_index))

    layoutOptions = tech.load_layout_options

    layoutOptions.lefdef_config.macro_resolution_mode = 1
    layoutOptions.lefdef_config.via_cellname_prefix = "VIA_"
    pathed_files = set()
    for lef_file in layoutOptions.lefdef_config.lef_files:
        if foundry_lef and not os.path.isabs(lef_file):
            lef_file = os.path.join(os.path.dirname(foundry_lef), lef_file)
        lef_file = os.path.abspath(lef_file)
        if os.path.isfile(lef_file):
            pathed_files.add(os.path.abspath(lef_file))

    for lef in lefs:
        pathed_files.add(os.path.abspath(lef))

    layoutOptions.lefdef_config.lef_files = list(pathed_files)
    layoutOptions.lefdef_config.read_lef_with_def = False
    layoutOptions.lefdef_config.dbu = tech.dbu

    layoutOptions.lefdef_config.produce_fills = True

    for lef_file in layoutOptions.lefdef_config.lef_files:
        print(f"[INFO] LEF file: {lef_file}")

    # Set layer properties
    layer_properties = tech.layer_properties_file
    if layer_properties:
        layer_properties = layer_properties[0]
        if not os.path.isabs(layer_properties):
            layer_properties = os.path.abspath(os.path.join(os.path.dirname(tech_file),
                                                            layer_properties))
    if os.path.exists(local_files['lyp']):
        layer_properties = os.path.abspath(local_files['lyp'])
    elif schema.valid('pdk', sc_pdk, 'display', 'klayout', sc_stackup):
        pdk_layer_props = schema.get('pdk', sc_pdk, 'display', 'klayout', sc_stackup)
        if pdk_layer_props:
            layer_properties = pdk_layer_props[0]

    if layer_properties and os.path.exists(layer_properties):
        tech.layer_properties_file = layer_properties
        print(f"[INFO] Layer properties: {layer_properties}")

    # Set layer map
    map_file = layoutOptions.lefdef_config.map_file
    if map_file:
        map_file = map_file[0]
        if not os.path.isabs(map_file):
            map_file = os.path.abspath(os.path.join(os.path.dirname(tech_file),
                                                    map_file))
    for s in get_streams(schema):
        if schema.valid('pdk', sc_pdk, 'layermap', 'klayout', 'def', s, sc_stackup):
            map_file = schema.get('pdk', sc_pdk, 'layermap', 'klayout', 'def', s, sc_stackup)
            if map_file:
                map_file = map_file[0]
                break

    if map_file and os.path.exists(map_file):
        layoutOptions.lefdef_config.map_file = map_file
        print(f"[INFO] Layer map: {map_file}")

    tech.load_layout_options = layoutOptions

    return tech


def save_technology(design, tech):
    tech.default_base_path = '.'
    tech.base_path = '.'

    if tech.layer_properties_file:
        layer_file = f'{design}.lyp'
        shutil.copyfile(tech.layer_properties_file,
                        f'outputs/{layer_file}')
        tech.layer_properties_file = layer_file

    tech.save(f'outputs/{design}.lyt')


def get_write_options(filename, timestamps):
    write_options = pya.SaveLayoutOptions()
    write_options.gds2_write_timestamps = timestamps
    write_options.set_format_from_filename(filename)

    return write_options


def get_schema(manifest):
    scroot = os.path.join(os.path.dirname(__file__), '..', '..')
    module_name = 'schema'
    schema_base = os.path.join(scroot, module_name, '__init__.py')
    spec = importlib_util.spec_from_file_location(module_name, schema_base)
    module = importlib_util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    # Return schema
    return module.Schema(manifest=manifest)



# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.

# NOTE: this file cannot rely on any third-party dependencies, including other
# SC dependencies outside of its directory, since it may be used by tool drivers
# that have isolated Python environments.

import os
import re
import sys

PACKAGE_ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')


def escape_val_tcl(val, typestr):
    '''Recursive helper function for converting Python values to safe TCL
    values, based on the SC type string.'''
    if val is None:
        return ''
    elif typestr.startswith('('):
        # Recurse into each item of tuple
        subtypes = typestr.strip('()').split(',')
        valstr = ' '.join(escape_val_tcl(v, subtype.strip())
                          for v, subtype in zip(val, subtypes))
        return f'[list {valstr}]'
    elif typestr.startswith('['):
        # Recurse into each item of list
        subtype = typestr.strip('[]')
        valstr = ' '.join(escape_val_tcl(v, subtype) for v in val)
        return f'[list {valstr}]'
    elif typestr == 'bool':
        return 'true' if val else 'false'
    elif typestr in ('str', 'enum'):
        # Escape string by surrounding it with "" and escaping the few
        # special characters that still get considered inside "". We don't
        # use {}, since this requires adding permanent backslashes to any
        # curly braces inside the string.
        # Source: https://www.tcl.tk/man/tcl8.4/TclCmd/Tcl.html (section [4] on)
        escaped_val = (val.replace('\\', '\\\\')  # escape '\' to avoid backslash substitution
                                                  # (do this first, since other replaces insert '\')
                          .replace('[', '\\[')    # escape '[' to avoid command substitution
                          .replace('$', '\\$')    # escape '$' to avoid variable substitution
                          .replace('"', '\\"'))   # escape '"' to avoid string terminating early
        return '"' + escaped_val + '"'
    elif typestr in ('file', 'dir'):
        # Replace $VAR with $env(VAR) for tcl
        val = re.sub(r'\$(\w+)', r'$env(\1)', val)
        # Same escapes as applied to string, minus $ (since we want to resolve env vars).
        escaped_val = (val.replace('\\', '\\\\')  # escape '\' to avoid backslash substitution
                                                  # (do this first, since other replaces insert '\')
                          .replace('[', '\\[')    # escape '[' to avoid command substitution
                          .replace('"', '\\"'))   # escape '"' to avoid string terminating early
        return '"' + escaped_val + '"'
    elif typestr in ('int', 'float'):
        # floats/ints just become strings
        return str(val)
    else:
        raise TypeError(f'{typestr} is not a supported type')


def trim(docstring):
    '''Helper function for cleaning up indentation of docstring.

    This is important for properly parsing complex RST in our docs.

    Source:
    https://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation'''
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def translate_loglevel(level):
    return level.upper()



import os
import shutil

from siliconcompiler.tools.klayout import klayout
from siliconcompiler.tools.klayout.klayout import setup as setup_tool
from siliconcompiler.tools._common import find_incoming_ext, get_tool_task


def make_docs(chip):
    klayout.make_docs(chip)
    chip.set('tool', 'klayout', 'task', 'show', 'var', 'show_filepath', '<path>')


def general_gui_setup(chip, task, exit, require_input=True):
    # Generic tool setup.
    setup_tool(chip)

    tool = 'klayout'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    clobber = False

    script = 'klayout_show.py'
    chip.set('tool', tool, 'task', task, 'script', script, step=step, index=index, clobber=clobber)

    pdk = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')
    if chip.valid('pdk', pdk, 'var', 'klayout', 'hide_layers', stackup):
        layers_to_hide = chip.get('pdk', pdk, 'var', 'klayout', 'hide_layers', stackup)
        chip.add('tool', tool, 'task', task, 'var', 'hide_layers', layers_to_hide,
                 step=step, index=index)

    if chip.valid('tool', tool, 'task', task, 'var', 'show_filepath') and \
       chip.get('tool', tool, 'task', task, 'var', 'show_filepath', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'show_filepath']),
                 step=step, index=index)
    elif require_input:
        incoming_ext = find_incoming_ext(chip, ('gds', 'oas', 'def'), 'def')
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'show_filetype']),
                 step=step, index=index)
        chip.set('tool', tool, 'task', task, 'var', 'show_filetype', incoming_ext,
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'input', f'{chip.top()}.{incoming_ext}',
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'show_exit', "true" if exit else "false",
             step=step, index=index, clobber=False)

    # Help
    chip.set('tool', tool, 'task', task, 'var', 'hide_layers',
             'List of layers to hide',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'show_filepath',
             'File to open',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'show_filetype',
             'File type to look for in the inputs',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'show_exit',
             'true/false: true will cause kLayout to exit when complete',
             field='help')


def setup(chip):
    '''
    Show a layout in kLayout
    '''

    # Generic tool setup.
    setup_tool(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    clobber = False

    chip.set('tool', tool, 'task', task, 'threads', 1, step=step, index=index, clobber=clobber)

    general_gui_setup(chip, task, False)

    option = ['-nc', '-rm']
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index, clobber=clobber)


def pre_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if chip.valid('tool', tool, 'task', task, 'var', 'show_filepath') and \
       chip.get('tool', tool, 'task', task, 'var', 'show_filepath', step=step, index=index):
        show_file = chip.get('tool', tool, 'task', task, 'var', 'show_filepath',
                             step=step, index=index)[0]

        rel_path = os.path.dirname(show_file)
        for ext in ('lyt', 'lyp'):
            ext_file = os.path.join(rel_path, f'{chip.top()}.{ext}')
            if ext_file and os.path.exists(ext_file):
                shutil.copy2(ext_file, f"inputs/{chip.top()}.{ext}")



from siliconcompiler.tools.vpr.vpr import parse_version as vpr_parse_version
from siliconcompiler.tools.vpr.vpr import normalize_version as vpr_normalize_version
from siliconcompiler.tools.vpr.vpr import add_tool_requirements as add_vpr_requirements

'''
Generate a `FSAM <https://github.com/chipsalliance/fasm>`_ file from the output of
`VPR <https://github.com/verilog-to-routing/vtr-verilog-to-routing>`_

Documentation: https://docs.verilogtorouting.org/en/latest/utils/fasm/

Sources: https://github.com/verilog-to-routing/vtr-verilog-to-routing/tree/master/utils/fasm
'''


######################################################################
# Make Docs
######################################################################
def make_docs(chip):
    from siliconcompiler.tools.genfasm.bitstream import setup
    setup(chip)
    return chip


def setup(chip):
    chip.set('tool', 'genfasm', 'exe', 'genfasm', clobber=False)
    chip.set('tool', 'genfasm', 'vswitch', '--version')
    chip.set('tool', 'genfasm', 'version', '>=8.1.0', clobber=False)

    add_tool_requirements(chip)


def add_tool_requirements(chip):
    add_vpr_requirements(chip)


def parse_version(chip):
    return vpr_parse_version(chip)


def normalize_version(chip):
    return vpr_normalize_version(chip)


##################################################
if __name__ == "__main__":
    chip = make_docs()
    chip.write_manifest("genfasm.json")



import os
import siliconcompiler
from siliconcompiler.utils import register_sc_data_source


####################################################
# Setup for vpr_example Family FPGAs
####################################################
def setup():
    '''
    The vpr_example FPGA family is a set of
    open source architectures used as illustrative
    examples for academic FPGA architectures.  They
    are based on numerous examples furnished over the
    the years by the University of Toronto with different
    distributions of VPR

    For more information about VPR and its architecture models,
    see Murray et. al, "VTR 8: High Performance CAD and Customizable
    FPGA Architecture Modelling", ACM Trans. Reconfigurable Technol.
    Syst., 2020, https://www.eecg.utoronto.ca/~kmurray/vtr/vtr8_trets.pdf
    '''

    vendor = 'N/A'

    flow_root = os.path.join('examples', 'fpga_flow')

    lut_size = '4'

    all_fpgas = []

    all_part_names = [
        'example_arch_X005Y005',
        'example_arch_X008Y008',
        'example_arch_X014Y014',
        'example_arch_X030Y030',
    ]

    # Settings common to all parts in family
    for part_name in all_part_names:
        fpga = siliconcompiler.FPGA(part_name, package='siliconcompiler_data')
        register_sc_data_source(fpga)

        fpga.set('fpga', part_name, 'vendor', vendor)

        # Part name is specified per architecture file.  Device code specifies
        # which <fixed_layout> name to use when running VPR.  These examples
        # use the following names:
        if (part_name == 'example_arch_X005Y005'):
            fpga.set('fpga', part_name, 'var', 'vpr_device_code', 'fpga_beta')
        else:
            fpga.set('fpga', part_name, 'var', 'vpr_device_code', part_name)

        fpga.set('fpga', part_name, 'lutsize', lut_size)

        arch_root = os.path.join(flow_root, 'arch', part_name)
        fpga.set('fpga', part_name, 'file', 'archfile', os.path.join(arch_root, f'{part_name}.xml'))

        fpga.set('fpga', part_name, 'var', 'vpr_clock_model', 'ideal')

        if (part_name == 'example_arch_X005Y005'):
            arch_root = os.path.join(flow_root, 'arch', part_name)
            fpga.set('fpga', part_name, 'file', 'graphfile',
                     os.path.join(arch_root, 'example_arch_X005Y005_rr_graph.xml'))
            fpga.set('fpga', part_name, 'var', 'channelwidth', '32')

        if (part_name == 'example_arch_X008Y008'):
            # No RR graph for this architecture to support testing
            fpga.set('fpga', part_name, 'var', 'channelwidth', '32')

        if ((part_name == 'example_arch_X014Y014') or (part_name == 'example_arch_X030Y030')):

            techlib_root = os.path.join(flow_root, 'techlib')

            if (part_name == 'example_arch_X014Y014'):
                fpga.set('fpga', part_name, 'file', 'constraints_map',
                         os.path.join(arch_root, f'{part_name}_constraint_map.json'))

            fpga.set('fpga', part_name, 'var', 'channelwidth', '80')
            fpga.add('fpga', part_name, 'var', 'feature_set', 'async_set')
            fpga.add('fpga', part_name, 'var', 'feature_set', 'async_reset')
            fpga.add('fpga', part_name, 'var', 'feature_set', 'enable')
            fpga.add('fpga', part_name, 'file', 'yosys_flop_techmap',
                     os.path.join(techlib_root, 'example_arch_techmap_flops.v'))

            fpga.add('fpga', part_name, 'file', 'yosys_dsp_techmap',
                     os.path.join(techlib_root, 'example_arch_techmap_dsp.v'))

            fpga.add('fpga', part_name, 'file', 'yosys_extractlib',
                     os.path.join(techlib_root, 'example_arch_techmap_dsp_extract.v'))

            # The same library used for the extraction pass can also be used to
            # define macros that can be passed through synthesis, specify that here
            fpga.add('fpga', part_name, 'file', 'yosys_macrolib',
                     os.path.join(techlib_root, 'example_arch_techmap_dsp_extract.v'))

            fpga.add('fpga', part_name, 'var', 'yosys_dsp_options', 'DSP_A_MAXWIDTH=18')
            fpga.add('fpga', part_name, 'var', 'yosys_dsp_options', 'DSP_B_MAXWIDTH=18')
            fpga.add('fpga', part_name, 'var', 'yosys_dsp_options', 'DSP_A_MINWIDTH=2')
            fpga.add('fpga', part_name, 'var', 'yosys_dsp_options', 'DSP_B_MINWIDTH=2')
            fpga.add('fpga', part_name, 'var', 'yosys_dsp_options', 'DSP_NAME=_dsp_block_')

            fpga.add('fpga', part_name, 'file', 'yosys_memory_techmap',
                     os.path.join(techlib_root, 'example_arch_techmap_bram.v'))
            fpga.add('fpga', part_name, 'file', 'yosys_memory_libmap',
                     os.path.join(techlib_root, 'example_arch_bram_memory_map.txt'))

        all_fpgas.append(fpga)

    return all_fpgas


#########################
if __name__ == "__main__":
    for fpga in setup(siliconcompiler.Chip('<fpga>')):
        fpga.write_manifest(f'{fpga.design}.json')



import os
from siliconcompiler.tools._common import \
    add_require_input, add_frontend_requires, get_input_files, get_frontend_options, \
    get_tool_task


def setup(chip):
    '''
    Compile the input verilog into a vvp file that can be simulated.
    '''

    # If the 'lock' bit is set, don't reconfigure.
    tool = 'icarus'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Standard Setup
    chip.set('tool', tool, 'exe', 'iverilog')
    chip.set('tool', tool, 'vswitch', '-V')
    chip.set('tool', tool, 'version', '>=10.3', clobber=False)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'verilog_generation',
             'Select Verilog language generation for Icarus to use. Legal values are '
             '"1995", "2001", "2001-noconfig", "2005", "2005-sv", "2009", or "2012". '
             'See the corresponding "-g" flags in the Icarus manual for more information.',
             field='help')

    add_require_input(chip, 'input', 'rtl', 'netlist')
    add_require_input(chip, 'input', 'rtl', 'verilog')
    add_require_input(chip, 'input', 'rtl', 'systemverilog')
    add_require_input(chip, 'input', 'cmdfile', 'f')
    add_frontend_requires(chip, ['ydir', 'vlib', 'idir', 'define', 'libext'])

    design = chip.top()
    chip.add('tool', tool, 'task', task, 'output', f'{design}.vvp', step=step, index=index)


################################
#  Custom runtime options
################################
def runtime_options(chip):

    ''' Custom runtime options, returns list of command line options.
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    cmdlist = []

    design = chip.top()
    cmdlist = ['-o', f'outputs/{design}.vvp']
    cmdlist += ['-s', chip.top()]

    verilog_gen = chip.get('tool', tool, 'task', task, 'var', 'verilog_generation',
                           step=step, index=index)
    if verilog_gen:
        cmdlist.append(f'-g{verilog_gen[0]}')

    opts = get_frontend_options(chip, ['ydir', 'vlib', 'idir', 'define', 'libext'])

    for libext in opts['libext']:
        cmdlist.append(f'-Y.{libext}')

    # source files
    for value in opts['ydir']:
        cmdlist.append('-y ' + value)
    for value in opts['vlib']:
        cmdlist.append('-v ' + value)
    for value in opts['idir']:
        cmdlist.append('-I' + value)
    for value in opts['define']:
        cmdlist.append('-D' + value)
    for value in get_input_files(chip, 'input', 'cmdfile', 'f'):
        cmdlist.append('-f ' + value)
    for value in get_input_files(chip, 'input', 'rtl', 'netlist'):
        cmdlist.append(value)
    for value in get_input_files(chip, 'input', 'rtl', 'verilog'):
        cmdlist.append(value)
    for value in get_input_files(chip, 'input', 'rtl', 'systemverilog'):
        cmdlist.append(value)

    return cmdlist



import siliconcompiler
from lambdapdk.ihp130 import setup


#########################
if __name__ == "__main__":
    pdk = setup(siliconcompiler.Chip('<pdk>'))
    pdk.write_manifest(f'{pdk.top()}.json')



'''
Lint system verilog
'''
from siliconcompiler.tools import slang
from siliconcompiler.tools._common import \
    add_require_input, add_frontend_requires, get_frontend_options, get_input_files, get_tool_task
import os


def setup(chip):
    slang.setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             clobber=False, step=step, index=index)

    add_require_input(chip, 'input', 'rtl', 'verilog')
    add_require_input(chip, 'input', 'rtl', 'systemverilog')
    add_frontend_requires(chip, ['ydir', 'idir', 'vlib', 'libext', 'define', 'param'])


def runtime_options(chip):
    options = ["-lint-only"]

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    options.extend(['-j', str(chip.get('tool', tool, 'task', task, 'threads',
                              step=step, index=index))])

    opts = get_frontend_options(chip,
                                ['ydir',
                                 'idir',
                                 'vlib',
                                 'libext',
                                 'define',
                                 'param'])

    if opts['libext']:
        options.append(f'--libext {",".join(opts["libext"])}')

    #####################
    # Library directories
    #####################
    if opts['ydir']:
        options.append(f'-y {",".join(opts["ydir"])}')

    #####################
    # Library files
    #####################
    if opts['vlib']:
        options.append(f'-libfile {",".join(opts["vlib"])}')

    #####################
    # Include paths
    #####################
    if opts['idir']:
        options.append(f'--include-directory {",".join(opts["idir"])}')

    #######################
    # Variable Definitions
    #######################
    for value in opts['define']:
        options.append('-D ' + value)

    #######################
    # Command files
    #######################
    cmdfiles = get_input_files(chip, 'input', 'cmdfile', 'f')
    if cmdfiles:
        options.append(f'-F {",".join(cmdfiles)}')

    #######################
    # Sources
    #######################
    for value in get_input_files(chip, 'input', 'rtl', 'systemverilog'):
        options.append(value)
    for value in get_input_files(chip, 'input', 'rtl', 'verilog'):
        options.append(value)

    #######################
    # Top Module
    #######################
    options.append('--top ' + chip.top())

    ###############################
    # Parameters (top module only)
    ###############################
    # Set up user-provided parameters to ensure we elaborate the correct modules
    for param, value in opts['param']:
        options.append(f'-G {param}={value}')

    return options


def post_process(chip):
    slang.post_process(chip)



from docutils import nodes
import sphinx.addnodes

from siliconcompiler.schema import Schema


# Docutils helpers
def build_table(items, colwidths=None, colspec=None):
    '''Create table node.

    Args:
        - items (list of list of nodes): nested list of table contents
        - colwidths (list of nums): relative column widths (seems to affect HTML
            builder only)
        - colspec (str): LaTeX column spec for overriding Sphinx defaults

    Returns list of nodes, since there may be an associated TabularColumn node
    that hints at sizing in PDF output.
    '''
    if colwidths is None:
        colwidths = [1] * len(items[0])  # default to equal spacing
    else:
        assert len(colwidths) == len(items[0])

    return_nodes = []
    if colspec is not None:
        colspec_node = sphinx.addnodes.tabular_col_spec()
        colspec_node['spec'] = colspec
        return_nodes.append(colspec_node)

    table = nodes.table()
    table['classes'] = ['longtable']

    group = nodes.tgroup(cols=len(items[0]))
    table += group
    for colwidth in colwidths:
        group += nodes.colspec(colwidth=colwidth)

    body = nodes.tbody()
    group += body

    for row in items:
        row_node = nodes.row()
        body += row_node
        for col in row:
            entry = nodes.entry()
            row_node += entry
            entry += col

    return_nodes.append(table)

    return return_nodes


def build_section(text, key):
    sec = nodes.section(ids=[get_ref_id(key)])
    sec += nodes.title(text=text)
    return sec


def build_section_with_target(text, key, ctx):
    id = get_ref_id(key)
    target = nodes.target('', '', ids=[id], names=[id])
    sec = nodes.section(ids=[id])
    sec += nodes.title(text=text)

    # We don't need to add target node to hierarchy, just need to call this
    # function.
    ctx.note_explicit_target(target)

    return sec


def get_ref_id(key):
    return nodes.make_id(key + "-ref")


def para(text):
    return nodes.paragraph(text=text)


def code(text):
    return nodes.literal(text=text)


def literalblock(text):
    block = nodes.literal_block(text=text)
    block['language'] = 'none'
    return block


def strong(text):
    p = nodes.paragraph()
    p += nodes.strong(text=text)
    return p


def image(src, center=False):
    i = nodes.image()
    i['uri'] = '/' + src
    if center:
        i['align'] = 'center'
    return i


def link(url, text=None):
    if text is None:
        text = url
    return nodes.reference(internal=False, refuri=url, text=text)


def build_list(items, enumerated=False):
    if enumerated:
        list = nodes.enumerated_list()
    else:
        list = nodes.bullet_list()

    for item in items:
        docutils_item = nodes.list_item()
        docutils_item += item
        list += docutils_item

    return list


def keypath(key_path, refdoc, key_text=None):
    '''Helper function for displaying Schema keypaths.'''
    text_parts = []
    key_parts = []
    cfg = Schema().cfg
    for key in key_path:
        if list(cfg.keys()) != ['default']:
            text_parts.append(f"'{key}'")
            key_parts.append(key)
            try:
                cfg = cfg[key]
            except KeyError:
                raise ValueError(f'Invalid keypath {key_path}')
        else:
            cfg = cfg['default']
            if key.startswith('<') and key.endswith('>'):
                # Placeholder
                text_parts.append(key)
            else:
                # Fully-qualified
                text_parts.append(f"'{key}'")

    if 'help' not in cfg:
        # Not leaf
        text_parts.append('...')

    if key_text:
        text_parts = key_text
    text = f"[{', '.join(text_parts)}]"
    refid = get_ref_id('param-' + '-'.join(key_parts))

    opt = {'refdoc': refdoc,
           'refdomain': 'sc',
           'reftype': 'ref',
           'refexplicit': True,
           'refwarn': True}
    refnode = sphinx.addnodes.pending_xref('keypath', **opt)
    refnode['reftarget'] = refid
    refnode += code(text)

    return refnode



from siliconcompiler.report.dashboard.layouts import vertical_flowgraph
from siliconcompiler.report.dashboard.layouts import vertical_flowgraph_sac_tabs
from siliconcompiler.report.dashboard.layouts import vertical_flowgraph_node_tab

__LAYOUTS = {
    "vertical_flowgraph": vertical_flowgraph.layout,
    "vertical_flowgraph_sac_tabs": vertical_flowgraph_sac_tabs.layout,
    "vertical_flowgraph_node_tab": vertical_flowgraph_node_tab.layout
}


def get_all_layouts():
    return list(__LAYOUTS.keys())


def get_layout(name):
    if name not in __LAYOUTS:
        raise ValueError(f"{name} is not a layout")

    return __LAYOUTS[name]
